from __future__ import annotations

import pathlib

import pytest
from lsprotocol import types
from pytest_lsp import LanguageClient

EXPECTED = {
    "ref",
    "doc",
    "option",
    "func",
    "class",
    "c:macro",
    "c:func",
    "py:func",
    "py:class",
    "std:ref",
    "std:doc",
}

UNEXPECTED = {
    "macro",
    "restructuredtext-unimplemented-role",
}


LOCAL_PY_CLASSES = {
    "counters.pattern.PatternCounter",
    "counters.pattern.NoMatchesError",
}
PYTHON_PY_CLASSES = {"logging.Filter", "http.server.HTTPServer"}
SPHINX_PY_CLASSES = {"sphinx.addnodes.desc"}


@pytest.mark.parametrize(
    "text, expected, unexpected",
    [
        ("::", None, None),
        (":", EXPECTED, UNEXPECTED),
        (":r", EXPECTED, UNEXPECTED),
        (":c:func", EXPECTED, UNEXPECTED),
        (":c:func: ", None, None),
        ("  ::", None, None),
        ("  :", EXPECTED, UNEXPECTED),
        ("  :r", EXPECTED, UNEXPECTED),
        ("  :c:func", EXPECTED, UNEXPECTED),
        ("  :c:func: ", None, None),
        ("(:", EXPECTED, UNEXPECTED),
        ("(:r", EXPECTED, UNEXPECTED),
        ("(:c:func", EXPECTED, UNEXPECTED),
    ],
)
@pytest.mark.asyncio(loop_scope="session")
async def test_rst_role_completions(
    client: LanguageClient,
    uri_for,
    text: str,
    expected: set[str] | None,
    unexpected: set[str] | None,
):
    """Ensure that the language server can offer role completions in rst documents."""
    test_uri = uri_for("workspaces", "demo", "rst", "roles.rst")

    uri = str(test_uri)
    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()
    linum = contents.splitlines().index(".. Add your reference here...")

    # Open the file
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=uri,
                language_id="restructuredtext",
                version=1,
                text=contents,
            )
        )
    )

    # Write some text
    #
    # This should replace the '.. Add your note here...' comment in
    # 'demo/rst/directives.rst' with the provided text
    client.text_document_did_change(
        types.DidChangeTextDocumentParams(
            text_document=types.VersionedTextDocumentIdentifier(uri=uri, version=2),
            content_changes=[
                types.TextDocumentContentChangePartial(
                    text=text,
                    range=types.Range(
                        start=types.Position(line=linum, character=0),
                        end=types.Position(line=linum + 1, character=0),
                    ),
                )
            ],
        )
    )

    # Make the completion request
    results = await client.text_document_completion_async(
        types.CompletionParams(
            text_document=types.TextDocumentIdentifier(uri=uri),
            position=types.Position(line=linum, character=len(text)),
        )
    )

    # Close the document - without saving!
    client.text_document_did_close(
        types.DidCloseTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=uri)
        )
    )

    if expected is None:
        assert results is None
    else:
        items = {item.label for item in results.items}
        unexpected = unexpected or set()

        assert expected == items & expected
        assert set() == items & unexpected


@pytest.mark.parametrize(
    "text, expected, unexpected",
    [
        (":ref:`", {"genindex", "modindex", "rst-roles-completion"}, set()),
        (":std:ref:`", {"genindex", "modindex", "rst-roles-completion"}, set()),
        (":doc:`", {"demo_myst", "demo_rst", "rst/domains/python"}, set()),
        (":std:doc:`", {"demo_myst", "demo_rst", "rst/domains/python"}, set()),
        (
            ":download:`./",
            {"roles.rst", "directives.rst", "domains", "domains.rst"},
            set(),
        ),
        (":download:`../", {"conf.py", "demo_rst.rst", "demo_myst.md"}, set()),
        (
            ":class:`",
            LOCAL_PY_CLASSES,
            PYTHON_PY_CLASSES | SPHINX_PY_CLASSES,
        ),
        (
            ":py:class:`",
            LOCAL_PY_CLASSES,
            PYTHON_PY_CLASSES | SPHINX_PY_CLASSES,
        ),
        (
            ":external:py:class:`",
            PYTHON_PY_CLASSES | SPHINX_PY_CLASSES,
            LOCAL_PY_CLASSES,
        ),
        (
            ":external+python:py:class:`",
            PYTHON_PY_CLASSES,
            LOCAL_PY_CLASSES | SPHINX_PY_CLASSES,
        ),
        (
            ":external+sphinx:py:class:`",
            SPHINX_PY_CLASSES,
            LOCAL_PY_CLASSES | PYTHON_PY_CLASSES,
        ),
        (":func:`", {"counters.pattern.count_numbers"}, set()),
        (":py:func:`", {"counters.pattern.count_numbers"}, set()),
    ],
)
@pytest.mark.asyncio(loop_scope="session")
async def test_rst_role_target_completions(
    client: LanguageClient,
    uri_for,
    text: str,
    expected: set[str] | None,
    unexpected: set[str] | None,
):
    """Ensure that the language server can offer role target completions in rst
    documents."""
    test_uri = uri_for("workspaces", "demo", "rst", "roles.rst")

    uri = str(test_uri)
    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()
    linum = contents.splitlines().index(".. Add your reference here...")

    # Open the file
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=uri,
                language_id="restructuredtext",
                version=1,
                text=contents,
            )
        )
    )

    # Write some text
    #
    # This should replace the '.. Add your note here...' comment in
    # 'demo/rst/directives.rst' with the provided text
    client.text_document_did_change(
        types.DidChangeTextDocumentParams(
            text_document=types.VersionedTextDocumentIdentifier(uri=uri, version=2),
            content_changes=[
                types.TextDocumentContentChangePartial(
                    text=text,
                    range=types.Range(
                        start=types.Position(line=linum, character=0),
                        end=types.Position(line=linum + 1, character=0),
                    ),
                )
            ],
        )
    )

    # Make the completion request
    results = await client.text_document_completion_async(
        types.CompletionParams(
            text_document=types.TextDocumentIdentifier(uri=uri),
            position=types.Position(line=linum, character=len(text)),
        )
    )

    # Close the document - without saving!
    client.text_document_did_close(
        types.DidCloseTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=uri)
        )
    )

    if expected is None:
        assert results is None
    else:
        items = {item.label for item in results.items}
        unexpected = unexpected or set()

        assert expected == items & expected
        assert set() == items & unexpected


@pytest.mark.parametrize(
    "text, expected, unexpected",
    [
        ("{", EXPECTED, UNEXPECTED),
        ("{r", EXPECTED, UNEXPECTED),
        ("{c:func", EXPECTED, UNEXPECTED),
        ("{c:func} ", None, None),
        ("  {", EXPECTED, UNEXPECTED),
        ("  {r", EXPECTED, UNEXPECTED),
        ("  {c:func", EXPECTED, UNEXPECTED),
        ("  {c:func} ", None, None),
        ("({", EXPECTED, UNEXPECTED),
        ("({r", EXPECTED, UNEXPECTED),
        ("({c:func", EXPECTED, UNEXPECTED),
    ],
)
@pytest.mark.asyncio(loop_scope="session")
async def test_myst_role_completions(
    client: LanguageClient,
    uri_for,
    text: str,
    expected: set[str] | None,
    unexpected: set[str] | None,
):
    """Ensure that the language server can offer completions in MyST documents."""
    test_uri = uri_for("workspaces", "demo", "myst", "roles.md")

    uri = str(test_uri)
    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()
    linum = contents.splitlines().index("% Add your reference here...")

    # Open the file
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=uri,
                language_id="markdown",
                version=1,
                text=contents,
            )
        )
    )

    # Write some text
    #
    # This should replace the '% Add your note here...' comment in
    # 'demo/myst/directives.md' with the provided text
    client.text_document_did_change(
        types.DidChangeTextDocumentParams(
            text_document=types.VersionedTextDocumentIdentifier(uri=uri, version=2),
            content_changes=[
                types.TextDocumentContentChangePartial(
                    text=text,
                    range=types.Range(
                        start=types.Position(line=linum, character=0),
                        end=types.Position(line=linum + 1, character=0),
                    ),
                )
            ],
        )
    )

    # Make the completion request
    results = await client.text_document_completion_async(
        types.CompletionParams(
            text_document=types.TextDocumentIdentifier(uri=uri),
            position=types.Position(line=linum, character=len(text)),
        )
    )

    # Close the document - without saving!
    client.text_document_did_close(
        types.DidCloseTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=uri)
        )
    )

    if expected is None:
        assert results is None
    else:
        items = {item.label for item in results.items}
        unexpected = unexpected or set()

        assert expected == items & expected
        assert set() == items & unexpected


@pytest.mark.parametrize(
    "text, expected, unexpected",
    [
        ("{ref}`", {"genindex", "modindex", "rst-roles-completion"}, set()),
        ("{std:ref}`", {"genindex", "modindex", "rst-roles-completion"}, set()),
        ("{doc}`", {"demo_myst", "demo_rst", "rst/domains/python"}, set()),
        ("{std:doc}`", {"demo_myst", "demo_rst", "rst/domains/python"}, set()),
        ("{download}`./", {"roles.md", "directives.md"}, set()),
        ("{download}`../", {"conf.py", "demo_rst.rst", "demo_myst.md"}, set()),
        (
            "{class}`",
            LOCAL_PY_CLASSES,
            PYTHON_PY_CLASSES | SPHINX_PY_CLASSES,
        ),
        (
            "{py:class}`",
            LOCAL_PY_CLASSES,
            PYTHON_PY_CLASSES | SPHINX_PY_CLASSES,
        ),
        (
            "{external:py:class}`",
            PYTHON_PY_CLASSES | SPHINX_PY_CLASSES,
            LOCAL_PY_CLASSES,
        ),
        (
            "{external+sphinx:py:class}`",
            SPHINX_PY_CLASSES,
            LOCAL_PY_CLASSES | PYTHON_PY_CLASSES,
        ),
        (
            "{external+python:py:class}`",
            PYTHON_PY_CLASSES,
            LOCAL_PY_CLASSES | SPHINX_PY_CLASSES,
        ),
        ("{func}`", {"counters.pattern.count_numbers"}, set()),
        ("{py:func}`", {"counters.pattern.count_numbers"}, set()),
    ],
)
@pytest.mark.asyncio(loop_scope="session")
async def test_myst_role_target_completions(
    client: LanguageClient,
    uri_for,
    text: str,
    expected: set[str] | None,
    unexpected: set[str] | None,
):
    """Ensure that the language server can offer completions in MyST documents."""
    test_uri = uri_for("workspaces", "demo", "myst", "roles.md")

    uri = str(test_uri)
    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()
    linum = contents.splitlines().index("% Add your reference here...")

    # Open the file
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=uri,
                language_id="markdown",
                version=1,
                text=contents,
            )
        )
    )

    # Write some text
    #
    # This should replace the '% Add your note here...' comment in
    # 'demo/myst/directives.md' with the provided text
    client.text_document_did_change(
        types.DidChangeTextDocumentParams(
            text_document=types.VersionedTextDocumentIdentifier(uri=uri, version=2),
            content_changes=[
                types.TextDocumentContentChangePartial(
                    text=text,
                    range=types.Range(
                        start=types.Position(line=linum, character=0),
                        end=types.Position(line=linum + 1, character=0),
                    ),
                )
            ],
        )
    )

    # Make the completion request
    results = await client.text_document_completion_async(
        types.CompletionParams(
            text_document=types.TextDocumentIdentifier(uri=uri),
            position=types.Position(line=linum, character=len(text)),
        )
    )

    # Close the document - without saving!
    client.text_document_did_close(
        types.DidCloseTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=uri)
        )
    )

    if expected is None:
        assert results is None
    else:
        items = {item.label for item in results.items}
        unexpected = unexpected or set()

        assert expected == items & expected
        assert set() == items & unexpected


@pytest.mark.asyncio(loop_scope="session")
@pytest.mark.parametrize(
    "filename,expected",
    [
        (
            ["workspaces", "demo", "rst", "roles.rst"],
            [
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/restructuredtext/basics.html#rst-roles",
                    tooltip="Roles - Sphinx v",  # don't check for a precise version!
                    range=types.Range(
                        start=types.Position(line=3, character=121),
                        end=types.Position(line=3, character=130),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/referencing.html#role-doc",
                    tooltip="doc - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=16, character=45),
                        end=types.Position(line=16, character=48),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/referencing.html#role-download",
                    tooltip="download - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=17, character=46),
                        end=types.Position(line=17, character=54),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/domains/python.html#role-py-class",
                    tooltip="py:class - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=18, character=52),
                        end=types.Position(line=18, character=60),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/domains/python.html#role-py-func",
                    tooltip="py:func - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=18, character=83),
                        end=types.Position(line=18, character=90),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#role-external",
                    tooltip="external - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=19, character=84),
                        end=types.Position(line=19, character=92),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#ext-intersphinx",
                    tooltip="sphinx.ext.intersphinx – Link to other projects’ documentation - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=20, character=25),
                        end=types.Position(line=20, character=40),
                    ),
                ),
                types.DocumentLink(
                    target="${ROOT}/myst/roles.md",
                    range=types.Range(
                        start=types.Position(line=32, character=34),
                        end=types.Position(line=32, character=45),
                    ),
                ),
                types.DocumentLink(
                    target="${ROOT}/rst/roles.rst",
                    tooltip="Path exists",
                    range=types.Range(
                        start=types.Position(line=33, character=25),
                        end=types.Position(line=33, character=36),
                    ),
                ),
                types.DocumentLink(
                    target="https://docs.python.org/3/howto/logging.html#logging-exceptions",
                    tooltip="Exceptions raised during logging - Python v",
                    range=types.Range(
                        start=types.Position(line=34, character=49),
                        end=types.Position(line=34, character=67),
                    ),
                ),
                types.DocumentLink(
                    target="${ROOT}/rst/roles.rst",
                    tooltip="Path exists",
                    range=types.Range(
                        start=types.Position(line=46, character=25),
                        end=types.Position(line=46, character=36),
                    ),
                ),
                types.DocumentLink(
                    target="${ROOT}/myst/roles.md",
                    range=types.Range(
                        start=types.Position(line=47, character=18),
                        end=types.Position(line=47, character=29),
                    ),
                ),
            ],
        ),
        (
            ["workspaces", "demo", "myst", "roles.md"],
            [
                types.DocumentLink(
                    target="https://myst-parser.readthedocs.io/en/latest/syntax/roles-and-directives.html#syntax-roles",
                    tooltip="Roles - an in-line extension point - MyST Parser v",  # don't check for a precise version!
                    range=types.Range(
                        start=types.Position(line=2, character=121),
                        end=types.Position(line=2, character=133),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/referencing.html#role-doc",
                    tooltip="doc - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=13, character=45),
                        end=types.Position(line=13, character=48),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/referencing.html#role-download",
                    tooltip="download - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=14, character=46),
                        end=types.Position(line=14, character=54),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/domains/python.html#role-py-class",
                    tooltip="py:class - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=15, character=52),
                        end=types.Position(line=15, character=60),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/domains/python.html#role-py-func",
                    tooltip="py:func - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=15, character=83),
                        end=types.Position(line=15, character=90),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#role-external",
                    tooltip="external - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=16, character=84),
                        end=types.Position(line=16, character=92),
                    ),
                ),
                types.DocumentLink(
                    target="https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#ext-intersphinx",
                    tooltip="sphinx.ext.intersphinx – Link to other projects’ documentation - Sphinx v",
                    range=types.Range(
                        start=types.Position(line=17, character=25),
                        end=types.Position(line=17, character=40),
                    ),
                ),
                types.DocumentLink(
                    target="${ROOT}/rst/roles.rst",
                    range=types.Range(
                        start=types.Position(line=28, character=34),
                        end=types.Position(line=28, character=44),
                    ),
                ),
                types.DocumentLink(
                    target="${ROOT}/myst/roles.md",
                    tooltip="Path exists",
                    range=types.Range(
                        start=types.Position(line=29, character=25),
                        end=types.Position(line=29, character=35),
                    ),
                ),
                types.DocumentLink(
                    target="https://docs.python.org/3/howto/logging.html#logging-exceptions",
                    tooltip="Exceptions raised during logging - Python v",
                    range=types.Range(
                        start=types.Position(line=30, character=49),
                        end=types.Position(line=30, character=67),
                    ),
                ),
                types.DocumentLink(
                    target="${ROOT}/myst/roles.md",
                    tooltip="Path exists",
                    range=types.Range(
                        start=types.Position(line=40, character=25),
                        end=types.Position(line=40, character=35),
                    ),
                ),
                types.DocumentLink(
                    target="${ROOT}/myst/roles.md",
                    range=types.Range(
                        start=types.Position(line=41, character=18),
                        end=types.Position(line=41, character=29),
                    ),
                ),
            ],
        ),
    ],
)
async def test_role_document_links(
    client: LanguageClient,
    uri_for,
    filename: list[str],
    expected: list[types.DocumentLink],
):
    """Ensure that we handle ``textDocument/documentLink`` requests correctly."""

    root_uri = str(uri_for("workspaces", "demo"))
    test_uri = uri_for(*filename)

    links = await client.text_document_document_link_async(
        types.DocumentLinkParams(
            text_document=types.TextDocumentIdentifier(uri=str(test_uri))
        )
    )

    assert len(links) == len(expected)

    for link, actual in zip(expected, links, strict=False):
        assert link.range == actual.range

        target = link.target.replace("${ROOT}", root_uri)
        assert target == actual.target

        if link.tooltip is not None:
            assert actual.tooltip.startswith(link.tooltip)


@pytest.mark.asyncio(loop_scope="session")
@pytest.mark.parametrize(
    "filename,position,expected",
    [
        (
            ["workspaces", "demo", "rst", "roles.rst"],
            # Requests for the role itself should return nothing
            types.Position(line=46, character=18),
            None,
        ),
        (
            ["workspaces", "demo", "rst", "roles.rst"],
            types.Position(line=46, character=29),
            [
                types.Location(
                    uri="${ROOT}/rst/roles.rst",
                    range=types.Range(
                        start=types.Position(line=0, character=0),
                        end=types.Position(line=1, character=0),
                    ),
                )
            ],
        ),
        (
            ["workspaces", "demo", "rst", "roles.rst"],
            types.Position(line=47, character=22),
            [
                types.Location(
                    uri="${ROOT}/myst/roles.md",
                    range=types.Range(
                        start=types.Position(line=0, character=0),
                        end=types.Position(line=1, character=0),
                    ),
                )
            ],
        ),
        (
            ["workspaces", "demo", "rst", "roles.rst"],
            types.Position(line=48, character=28),
            [
                types.Location(
                    uri="${ROOT}/myst/roles.md",
                    range=types.Range(
                        start=types.Position(line=4, character=0),
                        end=types.Position(line=5, character=0),
                    ),
                )
            ],
        ),
        (
            ["workspaces", "demo", "rst", "roles.rst"],
            types.Position(line=49, character=38),
            [
                types.Location(
                    uri="${ROOT}/rst/domains/python.rst",
                    range=types.Range(
                        start=types.Position(line=52, character=0),
                        end=types.Position(line=53, character=0),
                    ),
                )
            ],
        ),
        (
            ["workspaces", "demo", "myst", "roles.md"],
            # Requests for the role itself should return nothing
            types.Position(line=40, character=19),
            None,
        ),
        (
            ["workspaces", "demo", "myst", "roles.md"],
            types.Position(line=40, character=31),
            [
                types.Location(
                    uri="${ROOT}/myst/roles.md",
                    range=types.Range(
                        start=types.Position(line=0, character=0),
                        end=types.Position(line=1, character=0),
                    ),
                )
            ],
        ),
        (
            ["workspaces", "demo", "myst", "roles.md"],
            types.Position(line=41, character=22),
            [
                types.Location(
                    uri="${ROOT}/myst/roles.md",
                    range=types.Range(
                        start=types.Position(line=0, character=0),
                        end=types.Position(line=1, character=0),
                    ),
                )
            ],
        ),
        (
            ["workspaces", "demo", "myst", "roles.md"],
            types.Position(line=42, character=28),
            [
                types.Location(
                    uri="${ROOT}/rst/roles.rst",
                    range=types.Range(
                        start=types.Position(line=5, character=0),
                        end=types.Position(line=6, character=0),
                    ),
                )
            ],
        ),
        (
            ["workspaces", "demo", "myst", "roles.md"],
            types.Position(line=43, character=28),
            [
                types.Location(
                    uri="${ROOT}/rst/domains/python.rst",
                    range=types.Range(
                        start=types.Position(line=52, character=0),
                        end=types.Position(line=53, character=0),
                    ),
                )
            ],
        ),
    ],
)
async def test_role_target_definitions(
    client: LanguageClient,
    uri_for,
    filename: list[str],
    position: types.Position,
    expected: list[types.Location] | None,
):
    """Ensure that we handle ``textDocument/definition`` requests correctly for
    role targets."""

    root_uri = str(uri_for("workspaces", "demo"))
    test_uri = uri_for(*filename)

    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()

    # Open the file - needed so that esbonio can correctly determine the language_id of
    # the document. Strictly speaking, you could probably consider the fact that this is
    # necessary to be a bug... but 99% of the time I'm fairly sure the client will have
    # done this before making the goto definition call.
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=str(test_uri),
                language_id="restructuredtext"
                if fpath.suffix == ".rst"
                else "markdown",
                version=1,
                text=contents,
            )
        )
    )

    definitions = await client.text_document_definition_async(
        types.DefinitionParams(
            text_document=types.TextDocumentIdentifier(uri=str(test_uri)),
            position=position,
        )
    )

    if expected is None:
        assert definitions is None

    else:
        assert len(definitions) == len(expected)

        location: types.Location
        for location, actual in zip(expected, definitions, strict=False):
            expected_uri = location.uri.replace("${ROOT}", root_uri)
            assert expected_uri == actual.uri
            assert location.range == actual.range


@pytest.mark.asyncio(loop_scope="session")
@pytest.mark.parametrize(
    "filename,position,expected",
    [
        (
            ["workspaces", "demo", "rst", "roles.rst"],
            # Requests for the role itself should return nothing
            types.Position(line=60, character=10),
            None,
        ),
        (
            ["workspaces", "demo", "myst", "roles.md"],
            # Requests for the role itself should return nothing
            types.Position(line=53, character=10),
            None,
        ),
        (
            ["workspaces", "demo", "rst", "roles.rst"],
            # Requests for the role itself should return nothing
            types.Position(line=60, character=26),
            types.Hover(
                contents=types.MarkupContent(
                    kind=types.MarkupKind.Markdown,
                    value="This counter implementation counts",
                ),
                range=types.Range(
                    start=types.Position(line=60, character=14),
                    end=types.Position(line=60, character=47),
                ),
            ),
        ),
        (
            ["workspaces", "demo", "myst", "roles.md"],
            # Requests for the role itself should return nothing
            types.Position(line=53, character=26),
            types.Hover(
                contents=types.MarkupContent(
                    kind=types.MarkupKind.Markdown,
                    value="This counter implementation counts",
                ),
                range=types.Range(
                    start=types.Position(line=53, character=14),
                    end=types.Position(line=53, character=47),
                ),
            ),
        ),
        (
            ["workspaces", "demo", "rst", "roles.rst"],
            # Requests for the role itself should return nothing
            types.Position(line=61, character=10),
            types.Hover(
                contents=types.MarkupContent(
                    kind=types.MarkupKind.Markdown,
                    value="Helper for creating a PatternCounter",
                ),
                range=types.Range(
                    start=types.Position(line=61, character=10),
                    end=types.Position(line=61, character=51),
                ),
            ),
        ),
        (
            ["workspaces", "demo", "myst", "roles.md"],
            # Requests for the role itself should return nothing
            types.Position(line=54, character=15),
            types.Hover(
                contents=types.MarkupContent(
                    kind=types.MarkupKind.Markdown,
                    value="Helper for creating a PatternCounter",
                ),
                range=types.Range(
                    start=types.Position(line=54, character=13),
                    end=types.Position(line=54, character=54),
                ),
            ),
        ),
        (
            ["workspaces", "demo", "rst", "roles.rst"],
            # Requests for the role itself should return nothing
            types.Position(line=62, character=43),
            types.Hover(
                contents=types.MarkupContent(
                    kind=types.MarkupKind.Markdown,
                    value="The default pattern used",
                ),
                range=types.Range(
                    start=types.Position(line=62, character=9),
                    end=types.Position(line=62, character=43),
                ),
            ),
        ),
        (
            ["workspaces", "demo", "myst", "roles.md"],
            # Requests for the role itself should return nothing
            types.Position(line=55, character=43),
            types.Hover(
                contents=types.MarkupContent(
                    kind=types.MarkupKind.Markdown,
                    value="The default pattern used",
                ),
                range=types.Range(
                    start=types.Position(line=55, character=12),
                    end=types.Position(line=55, character=46),
                ),
            ),
        ),
    ],
)
async def test_role_target_hover(
    client: LanguageClient,
    uri_for,
    filename: list[str],
    position: types.Position,
    expected: types.Hover | None,
):
    """Ensure that we handle ``textDocument/hover`` requests correctly for
    role targets."""

    test_uri = uri_for(*filename)

    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()

    # Open the file - needed so that esbonio can correctly determine the language_id of
    # the document. Strictly speaking, you could probably consider the fact that this is
    # necessary to be a bug... but 99% of the time I'm fairly sure the client will have
    # done this before making the hover call.
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=str(test_uri),
                language_id="restructuredtext"
                if fpath.suffix == ".rst"
                else "markdown",
                version=1,
                text=contents,
            )
        )
    )

    hover = await client.text_document_hover_async(
        types.HoverParams(
            text_document=types.TextDocumentIdentifier(uri=str(test_uri)),
            position=position,
        )
    )

    if expected is None:
        assert hover is None

    else:
        assert hover.contents.value.startswith(expected.contents.value)
        assert hover.contents.kind is expected.contents.kind
        assert hover.range == expected.range



from __future__ import annotations

import typing

import pytest

from esbonio.server.features.project_manager import Project

if typing.TYPE_CHECKING:
    from sphinx.application import Sphinx


@pytest.mark.asyncio
async def test_python_domain_discovery(app: Sphinx, project: Project):
    """Ensure that we can correctly index all the objects associated with the Python domain."""

    expected = set()
    domain = app.env.domains["py"]

    for name, _, objtype, docname, _, _ in domain.get_objects():
        expected.add((name, objtype, docname))

    actual = set()
    db = await project.get_db()
    cursor = await db.execute(
        "SELECT name, objtype, docname FROM objects "
        "WHERE domain = 'py' AND project IS NULL"
    )

    for item in await cursor.fetchall():
        actual.add(item)

    assert expected == actual


@pytest.mark.asyncio
async def test_std_domain_discovery(app: Sphinx, project: Project):
    """Ensure that we can correctly index all the objects associated with the Python domain."""

    expected = set()
    domain = app.env.domains["std"]

    for name, _, objtype, docname, _, _ in domain.get_objects():
        expected.add((name, objtype, docname))

    actual = set()
    db = await project.get_db()
    cursor = await db.execute(
        "SELECT name, objtype, docname FROM objects "
        "WHERE domain = 'std' AND project IS NULL"
    )

    for item in await cursor.fetchall():
        actual.add(item)

    assert expected == actual


@pytest.mark.parametrize(
    "projname, domain, objtype, expected",
    [
        ("python", "py", "class", ["logging.Filter"]),
        ("sphinx", "py", "class", ["sphinx.addnodes.desc"]),
    ],
)
@pytest.mark.asyncio
async def test_intersphinx_domain_discovery(
    project: Project, domain: str, objtype: str, projname: str, expected: list[str]
):
    """Ensure that we can correctly index all the objects associated with the Python
    domain in intersphinx projects."""

    db = await project.get_db()
    cursor = await db.execute(
        "SELECT name FROM objects WHERE domain = ? AND objtype = ? AND project = ?",
        (domain, objtype, projname),
    )

    actual = set()
    for (name,) in await cursor.fetchall():
        actual.add(name)

    for name in expected:
        assert name in actual



import inspect
from typing import Any

from docutils.parsers.rst import roles as docutils_roles

from .. import types
from ..app import Database
from ..app import Sphinx
from ..util import as_json

ROLES_TABLE = Database.Table(
    "roles",
    [
        Database.Column(name="name", dtype="TEXT"),
        Database.Column(name="implementation", dtype="TEXT"),
        Database.Column(name="location", dtype="JSON"),
        Database.Column(name="target_providers", dtype="JSON"),
    ],
)


def get_impl_name(role: Any) -> str:
    try:
        return f"{role.__module__}.{role.__name__}"
    except AttributeError:
        return f"{role.__module__}.{role.__class__.__name__}"


def get_impl_location(impl: Any) -> types.Location | None:
    """Get the implementation location of the given role"""

    try:
        if (filepath := inspect.getsourcefile(impl)) is None:
            return None

        uri = types.Uri.for_file(filepath).resolve()
        source, line = inspect.getsourcelines(impl)

        location = types.Location(
            uri=str(uri),
            range=types.Range(
                start=types.Position(line=line - 1, character=0),
                end=types.Position(line=line + len(source), character=0),
            ),
        )

        return location
    except Exception:
        # TODO: Log the error somewhere..
        return None


def index_roles(app: Sphinx):
    """Index all the roles that are available to this app."""

    roles: dict[str, types.Role] = {}

    # Process the roles registered through Sphinx
    for name, impl, providers in app.esbonio._roles:
        roles[name] = types.Role(name, get_impl_name(impl), target_providers=providers)

    # Look any remaining docutils provided roles
    found_roles = {
        **docutils_roles._roles,  # type: ignore[attr-defined]
        **docutils_roles._role_registry,  # type: ignore[attr-defined]
    }

    for name, role in found_roles.items():
        if role == docutils_roles.unimplemented_role or name in roles:
            continue

        roles[name] = types.Role(name, get_impl_name(role))

    _add_providers(app, roles)

    app.esbonio.db.ensure_table(ROLES_TABLE)
    app.esbonio.db.clear_table(ROLES_TABLE)
    app.esbonio.db.insert_values(
        ROLES_TABLE, [r.to_db(as_json) for r in roles.values()]
    )


def setup(app: Sphinx):
    # Ensure that this runs as late as possibile
    app.connect("builder-inited", index_roles, priority=999)


def _add_providers(app: Sphinx, roles: dict[str, types.Role]):
    """Add provider definitions to built in role types we know about."""

    filepath_provider = types.Role.TargetProvider("filepath", {"root": app.srcdir})

    for name in ["download"]:
        if (role := roles.get(name)) is not None:
            role.target_providers = [filepath_provider]



from __future__ import annotations

import pathlib

from esbonio import server
from esbonio.server import Uri

from .project import Project


class ProjectManager(server.LanguageFeature):
    """Responsible for managing project instances."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.projects: dict[str, Project] = {}
        """Holds active project instances"""

    async def shutdown(self, params: None):
        for project in self.projects.values():
            await project.close()

    def register_project(self, scope: str, dbpath: str | pathlib.Path):
        """Register a project."""
        self.logger.debug("Registered project for scope '%s': '%s'", scope, dbpath)
        self.projects[scope] = Project(dbpath, self.converter)

    def get_project(self, uri: Uri) -> Project | None:
        """Return the project instance for the given uri, if available"""
        scope = self.server.configuration.scope_for(uri)

        if (project := self.projects.get(scope, None)) is None:
            self.logger.debug("No applicable project for uri: %s", uri)
            return None

        return project



import logging
import pathlib
import sys

import pytest
import pytest_asyncio
from lsprotocol.types import WorkspaceFolder
from pygls.exceptions import JsonRpcInternalError
from pygls.workspace import Workspace

from esbonio.server.features.sphinx_manager.client import ClientState
from esbonio.server.features.sphinx_manager.client_subprocess import (
    SubprocessSphinxClient,
)
from esbonio.server.features.sphinx_manager.client_subprocess import (
    make_test_sphinx_client,
)
from esbonio.server.features.sphinx_manager.config import SphinxConfig
from esbonio.server.features.sphinx_manager.config import SubProcess

logger = logging.getLogger(__name__)
STATIC_DIR = (
    pathlib.Path(__file__).parent.parent.parent / "esbonio" / "sphinx_agent" / "static"
).resolve()


@pytest.mark.asyncio
async def test_build_includes_webview_js(client: SubprocessSphinxClient, uri_for):
    """Ensure that builds include the ``webview.js`` script."""

    out = client.build_uri
    assert out is not None

    webview_js = STATIC_DIR / "webview.js"
    assert webview_js.exists()

    webview_script = webview_js.read_text()
    assert "editor/scroll" in webview_script

    # Ensure the script is included in the page
    index_html = pathlib.Path(out / "index.html")
    assert webview_script in index_html.read_text()


@pytest.mark.asyncio
async def test_build_content_override(client: SubprocessSphinxClient, uri_for):
    """Ensure that we can override the contents of a given src file when
    required."""

    out = client.build_uri
    src = client.src_uri
    assert out is not None and src is not None

    await client.build()

    # Before
    expected = "Welcome to the demo documentation"
    index_html = pathlib.Path(out / "index.html")
    assert expected in index_html.read_text()

    await client.build(
        content_overrides={str(src / "index.rst"): "My Custom Title\n==============="}
    )

    # Ensure the override was applied
    expected = "My Custom Title"
    print(index_html.read_text())
    assert expected in index_html.read_text()


@pytest_asyncio.fixture(scope="module", loop_scope="module")
async def client_build_error(uri_for, tmp_path_factory):
    """A sphinx client that will error when a build is triggered."""
    build_dir = tmp_path_factory.mktemp("build")
    demo_workspace = uri_for("workspaces", "demo")
    test_uri = demo_workspace / "index.rst"

    workspace = Workspace(
        None,
        workspace_folders=[
            WorkspaceFolder(uri=str(demo_workspace), name="demo"),
        ],
    )

    conf_dir = uri_for("workspaces", "demo-error-build").fs_path
    config = SphinxConfig(
        python_command=SubProcess(command=[sys.executable]),
        build_command=[
            "sphinx-build",
            "-b",
            "html",
            "-c",
            str(conf_dir),
            demo_workspace.fs_path,
            str(build_dir),
        ],
    )
    resolved = config.resolve(test_uri, workspace, logger)
    assert resolved is not None

    sphinx_client = await make_test_sphinx_client(resolved)
    assert sphinx_client.state == ClientState.Running

    yield sphinx_client

    await sphinx_client.stop()


@pytest.mark.asyncio(loop_scope="module")
async def test_build_error(client_build_error: SubprocessSphinxClient):
    """Ensure that when a build error occurs, useful information is reported."""

    with pytest.raises(
        JsonRpcInternalError, match="sphinx-build failed:.*division by zero.*"
    ):
        await client_build_error.build()



"""Subprocess implementation of the ``SphinxClient`` protocol."""

from __future__ import annotations

import asyncio
import logging
import pathlib
import sys
import typing
from uuid import uuid4

import platformdirs
from pygls.client import JsonRPCClient
from pygls.protocol import JsonRPCProtocol

from esbonio.server import EventSource
from esbonio.server import Uri
from esbonio.sphinx_agent import types

from .client import ClientState
from .config import SphinxConfig

if typing.TYPE_CHECKING:
    from typing import Any

    from .client import SphinxClient
    from .manager import SphinxManager


sphinx_logger = logging.getLogger("sphinx")


class SphinxAgentProtocol(JsonRPCProtocol):
    """Describes the protocol spoken between the client below and the sphinx agent."""

    def get_message_type(self, method: str) -> Any | None:
        return types.METHOD_TO_MESSAGE_TYPE.get(method, None)

    def get_result_type(self, method: str) -> Any | None:
        return types.METHOD_TO_RESPONSE_TYPE.get(method, None)


class SubprocessSphinxClient(JsonRPCClient):
    """JSON-RPC client used to drive a Sphinx application instance hosted in
    a separate subprocess.

    See :mod:`esbonio.sphinx_agent` for the implementation of the server component.
    """

    def __init__(
        self,
        config: SphinxConfig,
        logger: logging.Logger | None = None,
        protocol_cls=SphinxAgentProtocol,
        *args,
        **kwargs,
    ):
        super().__init__(*args, protocol_cls=protocol_cls, **kwargs)  # type: ignore[misc]

        self.id = str(uuid4())
        """The client's id."""

        self.config = config
        """Configuration values."""

        self.logger = logger or logging.getLogger(__name__)
        """The logger instance to use."""

        self.sphinx_info: types.SphinxInfo | None = None
        """Information about the Sphinx application the client is connected to."""

        self.state: ClientState | None = None
        """The current state of the client."""

        self.exception: Exception | None = None
        """The most recently encountered exception (if any)"""

        self._events = EventSource(self.logger)
        """The sphinx client can emit events."""

        self._startup_task: asyncio.Task[Any] | None = None
        """The startup task."""

        self._stderr_forwarder: asyncio.Task[Any] | None = None
        """A task that forwards the server's stderr to the test process."""

    def __repr__(self):
        if self.state is None:
            return "SphinxClient<None>"

        if self.state == ClientState.Errored:
            return f"SphinxClient<{self.state.name}: {self.exception}>"

        state = self.state.name
        command = " ".join(self.config.build_command)
        return f"SphinxClient<{state}: {command}>"

    def __await__(self):
        """Makes the client await-able"""
        if self._startup_task is None:
            self._startup_task = asyncio.create_task(self.start())

        return self._startup_task.__await__()

    @property
    def converter(self):
        return self.protocol._converter

    @property
    def builder(self) -> str:
        """The sphinx application's builder name"""
        if self.sphinx_info is None:
            raise RuntimeError("sphinx_info is None, has the client been started?")

        return self.sphinx_info.builder_name

    @property
    def src_uri(self) -> Uri:
        """The src uri of the Sphinx application."""
        if self.sphinx_info is None:
            raise RuntimeError("sphinx_info is None, has the client been started?")

        return Uri.for_file(self.sphinx_info.src_dir)

    @property
    def conf_uri(self) -> Uri:
        """The conf uri of the Sphinx application."""
        if self.sphinx_info is None:
            raise RuntimeError("sphinx_info is None, has the client been started?")

        return Uri.for_file(self.sphinx_info.conf_dir)

    @property
    def db(self) -> pathlib.Path:
        """Connection to the associated database."""
        if self.sphinx_info is None:
            raise RuntimeError("sphinx_info is None, has the client been started?")

        return pathlib.Path(self.sphinx_info.dbpath)

    @property
    def build_uri(self) -> Uri:
        """The build uri of the Sphinx application."""
        if self.sphinx_info is None:
            raise RuntimeError("sphinx_info is None, has the client been started?")

        return Uri.for_file(self.sphinx_info.build_dir)

    def add_listener(self, event: str, handler):
        self._events.add_listener(event, handler)

    async def server_exit(self, server: asyncio.subprocess.Process):
        """Called when the sphinx agent process exits."""

        #   0: all good
        # -15: terminated
        if server.returncode not in {0, -15}:
            self.exception = RuntimeError(server.returncode)
            self._set_state(ClientState.Errored)
            self.logger.error(
                f"sphinx-agent process exited with code: {server.returncode}"
            )

            if server.stderr is not None:
                stderr = await server.stderr.read()
                self.logger.error("Stderr:\n%s", stderr.decode("utf8"))

        # TODO: Should the upstream base client be doing this?
        # Cancel any pending futures.
        for id_, fut in self.protocol._request_futures.items():
            message = "Cancelled" if fut.cancel() else "Unable to cancel"
            self.logger.debug(
                "%s future '%s' for pending request '%s'", message, fut, id_
            )

        if self.state != ClientState.Errored:
            self._set_state(ClientState.Exited)

    async def start_io(self, cmd: str, *args, **kwargs):
        await super().start_io(cmd, *args, **kwargs)

        # Forward the server's stderr to this process' stderr
        if self._server and self._server.stderr:
            self._stderr_forwarder = asyncio.create_task(forward_stderr(self._server))

    async def restart(self) -> SphinxClient:
        """Restart the client."""
        await self.stop()

        # We need to reset the client's stop event.
        self._stop_event.clear()

        self._set_state(ClientState.Restarting)
        return await self.start()

    async def start(self) -> SphinxClient:
        """Start the client."""

        # Only try starting once.
        if self.state not in {None, ClientState.Restarting}:
            return self

        try:
            self._set_state(ClientState.Starting)
            sphinx = self.config.sphinx_command

            self.logger.debug("Python command: %r", sphinx.command)
            await self.start_io(*sphinx.command, env=sphinx.env, cwd=sphinx.cwd)

            params = types.CreateApplicationParams(
                command=self.config.build_command,
                config_overrides=self.config.config_overrides,
                context={
                    "cacheDir": platformdirs.user_cache_dir("esbonio", "swyddfa"),
                },
            )
            self.sphinx_info = await self.protocol.send_request_async(
                "sphinx/createApp", params
            )

            self._set_state(ClientState.Running)
            return self
        except Exception as exc:
            self.logger.debug("Unable to start SphinxClient: %s", exc, exc_info=True)

            self.exception = exc
            self._set_state(ClientState.Errored)

            return self

    def _set_state(self, new_state: ClientState):
        """Change the state of the client."""
        old_state, self.state = self.state, new_state

        self.logger.debug("SphinxClient[%s]: %s -> %s", self.id, old_state, new_state)
        self._events.trigger("state-change", self, old_state, new_state)

    async def stop(self):
        """Stop the client."""

        self.protocol.notify("exit", None)

        # Give the agent a little time to close.
        await asyncio.sleep(0.5)

        if self._stderr_forwarder:
            self._stderr_forwarder.cancel()

        self.logger.debug(self._async_tasks)
        await super().stop()

    async def build(
        self,
        *,
        filenames: list[str] | None = None,
        force_all: bool = False,
        content_overrides: dict[str, str] | None = None,
    ) -> types.BuildResult:
        """Trigger a Sphinx build."""

        params = types.BuildParams(
            filenames=filenames or [],
            force_all=force_all,
            content_overrides=content_overrides or {},
        )

        self._set_state(ClientState.Building)
        try:
            result = await self.protocol.send_request_async("sphinx/build", params)
            self._set_state(ClientState.Running)

            return result
        except Exception as exc:
            self.exception = exc
            self._set_state(ClientState.Errored)

            raise


async def forward_stderr(server: asyncio.subprocess.Process):
    if server.stderr is None:
        return

    # EOF is signalled with an empty bytestring
    while (line := await server.stderr.readline()) != b"":
        sphinx_logger.info(line.decode().rstrip())


def make_subprocess_sphinx_client(
    manager: SphinxManager, config: SphinxConfig
) -> SphinxClient:
    """Factory function for creating a ``SubprocessSphinxClient`` instance.

    Parameters
    ----------
    manager
       The manager instance creating the client

    config
       The Sphinx configuration

    Returns
    -------
    SphinxClient
       The configured client
    """
    client = SubprocessSphinxClient(config, logger=manager.logger)

    @client.feature("window/logMessage")
    def _on_msg(ls: SubprocessSphinxClient, params):
        sphinx_logger.info(params.message)

    @client.feature("$/progress")
    def _on_progress(ls: SubprocessSphinxClient, params):
        manager.report_progress(ls, params)

    return client


def make_test_sphinx_client(config: SphinxConfig) -> SubprocessSphinxClient:
    """Factory function for creating a ``SubprocessSphinxClient`` instance
    to use for testing."""
    logger = logging.getLogger("sphinx_client")
    logger.setLevel(logging.INFO)

    client = SubprocessSphinxClient(config)

    @client.feature("window/logMessage")
    def _(params):
        print(params.message, file=sys.stderr)  # noqa: T201

    @client.feature("$/progress")
    def _on_progress(params):
        logger.info("%s", params)

    return client



from __future__ import annotations

import pathlib
import typing

from docutils import nodes
from docutils.transforms import Transform
from sphinx import addnodes
from sphinx import version_info

from ..log import source_to_uri_and_linum

if typing.TYPE_CHECKING:
    from sphinx.application import Sphinx


STATIC_DIR = (pathlib.Path(__file__).parent.parent / "static").resolve()
WEBVIEW_JS = (STATIC_DIR / "webview.js").read_text()
ALLOWED_MODULES = {"docutils.nodes", "sphinx.addnodes"}


def has_source(node):
    if isinstance(node, nodes.Text):
        return False

    # For some reason, including `toctreenode` causes Sphinx 5.x and 6.x to crash with a
    # cryptic error.
    #
    # AssertionError: Losing "classes" attribute
    #
    # Caused by this line:
    # https://github.com/sphinx-doc/sphinx/blob/ec993dda3690f260345133c47a4a0f6ef0b18493/sphinx/environment/__init__.py#L630
    if isinstance(node, addnodes.toctree) and version_info[0] < 7:
        return False

    # It's not only limited to `toctreenodes`!
    #
    # The identical error is thrown when using esbonio with the `ablog` extension
    # See: https://github.com/swyddfa/esbonio/issues/874
    #
    # I think for now, the safest approach is to only handle nodes defined by Sphinx or
    # docutils.
    if node.__module__ not in ALLOWED_MODULES:
        return False

    return (node.line or 0) > 0 and node.source is not None


class source_locations(nodes.General, nodes.Element):
    """Index of all known source locations."""


def visit_source_locations(self, node):
    source_index: dict[int, tuple[str, int]] = node["index"]

    self.body.append('<div id="esbonio-marker-index" style="display: none">\n')
    for idx, (uri, linum) in source_index.items():
        self.body.append(
            f'  <span data-id="{idx}" data-uri="{uri}" data-line="{linum}"></span>\n'
        )

    self.body.append("</div>")

    # Inline the JS code we need to enable sync scrolling.
    #
    # Yes this "bloats" every page in the generated docs, but is generally more robust
    # see: https://github.com/swyddfa/esbonio/issues/810
    #      https://github.com/swyddfa/esbonio/issues/942
    self.body.append(f"<script>{WEBVIEW_JS}</script>")


def depart_source_locations(self, node): ...


class SourceLocationTransform(Transform):
    """Add source location information to the doctree.

    Used to support features like synchronised scrolling.
    """

    default_priority = 500

    def apply(self, **kwargs):
        current_line = 0
        current_source = None

        source_index = {}
        source_nodes = self.document.traverse(condition=has_source)

        for idx, node in enumerate(source_nodes):
            node_line = node.line or 0

            if node_line > current_line or node.source != current_source:
                uri, linum = source_to_uri_and_linum(f"{node.source}:{node.line}")

                if uri is None or linum is None:
                    continue

                source_index[idx] = (str(uri), linum)
                node["classes"].extend(["esbonio-marker", f"esbonio-marker-{idx}"])  # type: ignore[index]

                # Use the source and line reported by docutils.
                # Just in case source_to_uri_and_linum doesn't handle things correctly
                current_line = node_line
                current_source = node.source

        self.document.children.append(source_locations("", index=source_index))


def setup(app: Sphinx):
    app.add_node(
        source_locations, html=(visit_source_locations, depart_source_locations)
    )
    app.add_transform(SourceLocationTransform)



from __future__ import annotations

import importlib.util
import json
import logging
import os
import pathlib
import re
import sys
from typing import Any

import attrs
from pygls import IS_WIN
from pygls.workspace import Workspace

from esbonio.server import Uri

VARIABLE = re.compile(r"\$\{([^}]+)\}")


def get_module_path(module: str) -> pathlib.Path | None:
    """Return the path to the directory containing the given module name.

    Parameters
    ----------
    module
       A valid Python module name e.g. ``esbonio.sphinx_agent``

    Returns
    -------
    pathlib.Path | None
       The path to the directory containing the given module.
       If ``None``, the module could not be found.
    """
    spec = importlib.util.find_spec(module)
    if spec is None:
        return None

    if spec.origin is None:
        return None

    # origin = .../esbonio/sphinx_agent/__init__.py
    agent = pathlib.Path(spec.origin)
    return agent.parent.parent


@attrs.define
class SubProcess:
    """Captures the information necessary to spawn the Sphinx agent subprocess"""

    command: list[str] = attrs.field(factory=list)
    """The command to invoke, plus any additional arguments."""

    env: dict[str, str] = attrs.field(factory=dict)
    """Additional environment variables to set for the process."""

    cwd: str = attrs.field(default="${scopeFsPath}")
    """The working directory to use."""

    def resolve(
        self, uri: Uri, workspace: Workspace, logger: logging.Logger
    ) -> SubProcess | None:
        """Resolve the configuration based on user provided values.

        Parameters
        ----------
        uri
           The uri of the file we are creating the sphinx agent instace for

        workspace
           The user's workspace

        logger
           The logger instance to use.

        Returns
        -------
        SubProcess | None
           The fully resolved config object to use.
           If ``None``, a valid configuration could not be created.
        """
        if (cwd := self._resolve_cwd(uri, workspace, logger)) is None:
            return None

        logger.debug("cwd: %s", cwd)
        if len(command := self._resolve_python(logger, cwd)) == 0:
            return None

        if (env := self._resolve_env(logger)) is None:
            return None

        return SubProcess(command=command, env=env, cwd=cwd)

    def _resolve_python(self, logger: logging.Logger, cwd: str) -> list[str]:
        """Return the python command to use when launching the sphinx agent.

        This could be as simple as the path to the python interpreter in a
        particular virtual environment or a complex command such as
        ``hatch -e docs run python``.

        If the user has not configured a python command, this will fallback to
        using ``sys.executable``.

        Parameters
        ----------
        logger
           The logger instance to use

        Returns
        -------
        list[str]
           The command to use when invoking python
        """
        if len(command := list(self.command)) == 0:
            logger.warning(
                "No pythonCommand configured! Reusing the server's environment."
            )
            return [sys.executable]

        command = [_resolve_variable(c, cwd) for c in command]
        return command

    def _resolve_env(self, logger: logging.Logger) -> dict[str, str] | None:
        """Construct the environment variables to set for the process.

        Using the ``PYTHONPATH`` environment variable, we can inject additional Python
        packages into the user's Python environment. This method locates the
        installation path of the sphinx agent and ensures it's added to the front
        of the ``PYTHONPATH`` variable.

        Parameters
        ----------
        logger
           The logger instance to use

        Returns
        -------
        dict[str, str]
           The environment variables to use.
        """

        if (sphinx_agent := get_module_path("esbonio.sphinx_agent")) is None:
            logger.error("Unable to locate the `esbonio.sphinx_agent` module")
            return None

        python_path: list[pathlib.Path | str] = [sphinx_agent]

        if len(pypath := self.env.get("PYTHONPATH", "")) > 0:
            python_path.append(pypath)

        env = {
            **self.env,
            "PYTHONUNBUFFERED": "1",
            "PYTHONPATH": os.pathsep.join(str(p) for p in python_path),
        }

        # Only log the environment variables set *before* copying in the wider
        # environment
        logger.debug("env: %s", json.dumps(env, indent=2))
        for envname, value in os.environ.items():
            # Don't pass any vars we've explictly set.
            if envname in env:
                continue

            env[envname] = value

        return env

    def _resolve_cwd(
        self, uri: Uri, workspace: Workspace, logger: logging.Logger
    ) -> str | None:
        """Determine the working directory from which to launch the Sphinx agent.

        Parameters
        ----------
        uri
           The uri of the file we are creating an agent instance for

        workspace
           The user's workspace.

        logger
           The logger instance to use.

        Returns
        -------
        str | None
           The working directory to launch the sphinx agent in.
           If ``None``, the working directory could not be determined.
        """
        logger.debug(f"{self.cwd}")
        if self.cwd and self.cwd != "${scopeFsPath}":
            return self.cwd

        candidates = [Uri.parse(f) for f in workspace.folders.keys()]

        if workspace.root_uri is not None:
            if (root_uri := Uri.parse(workspace.root_uri)) not in candidates:
                candidates.append(root_uri)

        for folder in candidates:
            if str(uri).startswith(str(folder)):
                if (cwd := folder.fs_path) is None:
                    logger.error(
                        "Unable to determine working directory from %r", folder
                    )
                    return None

                return cwd

        return None


@attrs.define
class SphinxConfig:
    """Configuration for the sphinx agent subprocess."""

    enable_dev_tools: bool = attrs.field(default=False)
    """Flag to enable dev tools."""

    python_command: SubProcess = attrs.field(factory=SubProcess)
    """The command to use when launching the python interpreter."""

    build_command: list[str] = attrs.field(factory=list)
    """The sphinx-build command to use."""

    config_overrides: dict[str, Any] = attrs.field(factory=dict)
    """Overrides to apply to Sphinx's configuration."""

    @property
    def sphinx_command(self) -> SubProcess:
        """Return the command definition necessary to launch the sphinx agent."""
        command: list[str] = []
        python = self.python_command

        if len(python.command) == 0:
            raise ValueError("No python environment configured")

        if self.enable_dev_tools:
            command.extend(["lsp-devtools", "agent", "--"])

        command.extend([*python.command, "-m", "sphinx_agent"])
        return SubProcess(command=command, env=python.env, cwd=python.cwd)

    def resolve(
        self,
        uri: Uri,
        workspace: Workspace,
        logger: logging.Logger,
    ) -> SphinxConfig | None:
        """Resolve the configuration based on user provided values.

        Parameters
        ----------
        uri
           The uri of the file we are creating the sphinx agent instace for

        workspace
           The user's workspace

        logger
           The logger instance to use.

        Returns
        -------
        SphinxConfig | None
           The fully resolved config object to use.
           If ``None``, a valid configuration could not be created.
        """
        python_command = self.python_command.resolve(uri, workspace, logger)
        if python_command is None:
            return None

        build_command = self._resolve_build_command(uri, logger)
        if len(build_command) == 0:
            return None

        logger.debug("Build command: %r", build_command)

        return SphinxConfig(
            enable_dev_tools=self.enable_dev_tools,
            config_overrides=self.config_overrides,
            python_command=python_command,
            build_command=build_command,
        )

    def _resolve_build_command(self, uri: Uri, logger: logging.Logger) -> list[str]:
        """Return the ``sphinx-build`` command to use.

        If no command is configured, this will attempt to guess the command to use based
        on the user's workspace.

        Parameters
        ----------
        uri
           The uri of the file we are creating the sphinx agent for.

        logger
           The logger instance to use.

        Returns
        -------
        List[str]
           The ``sphinx-build`` command to use.
           If empty, no build command could be determined.
        """

        if len(self.build_command) > 0:
            return self.build_command

        if (path := uri.fs_path) is None:
            return []

        # Search upwards from the given uri to see if we find something that looks like
        # a sphinx conf.py file.
        previous = None
        current = pathlib.Path(path)

        while previous != current:
            previous = current
            current = previous.parent

            conf_py = current / "conf.py"
            logger.debug("Trying path: %s", current)
            if conf_py.exists():
                return [
                    "sphinx-build",
                    "-M",
                    "dirhtml",
                    str(current),
                    "${defaultBuildDir}",
                ]

        return []


def _resolve_variable(arg: str, cwd: str) -> str:
    """Resolve the configuration variables in the given argument, if any.

    Parameters
    ----------
    arg
       The string containing the vatiables to resolve

    cwd
       The current working directory of the configuration.
    """

    if (match := VARIABLE.match(arg)) is None:
        return arg

    varname = match.group(1)

    if varname.startswith("venv:"):
        env = varname.replace("venv:", "")
        return _resolve_variable_venv(env, cwd)

    raise ValueError(f"Undefined variable: {varname!r}")


def _resolve_variable_venv(env: str, cwd: str) -> str:
    """Resolve the ``${venv:<path>}`` config variable"""
    if IS_WIN:
        envpath = pathlib.Path(env, "Scripts", "python.exe")
    else:
        envpath = pathlib.Path(env, "bin", "python")

    if envpath.is_absolute():
        return str(envpath)
    else:
        envpath = cwd / envpath
        # Can't call envpath.resolve() here as that will also follow symlinks which we
        # don't want to do
        # https://github.com/swyddfa/esbonio/issues/945
        return os.path.normpath(envpath)



from __future__ import annotations

import dataclasses
import hashlib
import inspect
import pathlib
import re
import sys
from typing import Any
from typing import Literal
from unittest import mock

from sphinx.application import Sphinx
from sphinx.cmd.build import main as sphinx_build

VARIABLE = re.compile(r"\$\{(\w+)\}")


@dataclasses.dataclass
class SphinxConfig:
    """Configuration values to pass to the Sphinx application instance."""

    src_dir: str
    """The directory containing the project's source."""

    conf_dir: str | None
    """The directory containing the project's ``conf.py``."""

    build_dir: str
    """The directory to write build outputs into."""

    builder_name: str
    """The currently used builder name."""

    doctree_dir: str
    """The directory to write doctrees into."""

    config_overrides: dict[str, Any] = dataclasses.field(default_factory=dict)
    """Any overrides to configuration values."""

    force_full_build: bool = dataclasses.field(default=False)
    """Force a full build on startup."""

    keep_going: bool = dataclasses.field(default=False)
    """Continue building when errors (from warnings) are encountered."""

    num_jobs: Literal["auto"] | int = dataclasses.field(default=1)
    """The number of jobs to use for parallel builds."""

    quiet: bool = dataclasses.field(default=False)
    """Hide standard Sphinx output messages"""

    silent: bool = dataclasses.field(default=False)
    """Hide all Sphinx output."""

    tags: list[str] = dataclasses.field(default_factory=list)
    """Tags to enable during a build."""

    verbosity: int = dataclasses.field(default=0)
    """The verbosity of Sphinx's output."""

    version: str | None = dataclasses.field(default=None)
    """Sphinx's version number."""

    warning_is_error: bool = dataclasses.field(default=False)
    """Treat any warning as an error"""

    @property
    def parallel(self) -> int:
        """The parsed value of the ``num_jobs`` field."""

        if self.num_jobs == "auto":
            import multiprocessing

            return multiprocessing.cpu_count()

        return self.num_jobs

    @classmethod
    def fromcli(cls, args: list[str]):
        """Return the ``SphinxConfig`` instance that's equivalent to the given arguments.

        Parameters
        ----------
        args
           The cli arguments you would normally pass to ``sphinx-build``

        Returns
        -------
        Optional[SphinxConfig]
           ``None`` if the arguments could not be parsed, otherwise the set configuration
           options derived from the sphinx build command.
        """

        if args[0] == "sphinx-build":
            args = args[1:]

        # The easiest way to handle this is to just call sphinx-build but with
        # the Sphinx app object patched out - then we just use all the args it
        # was given!
        with mock.patch("sphinx.cmd.build.Sphinx") as m_Sphinx:
            sphinx_build(args)

        if m_Sphinx.call_args is None:
            return None

        signature = inspect.signature(Sphinx)
        keys = signature.parameters.keys()

        values = m_Sphinx.call_args[0]
        sphinx_args = {k: v for k, v in zip(keys, values, strict=False)}

        # Sphinx 8.1 changed the way arguments are passed to the `Sphinx` class.
        # See: https://github.com/swyddfa/esbonio/issues/912
        if len(values) == 0:
            sphinx_args = m_Sphinx.call_args.kwargs

        if sphinx_args is None:
            return None

        return cls(
            src_dir=sphinx_args["srcdir"],
            conf_dir=sphinx_args["confdir"],
            build_dir=sphinx_args["outdir"],
            builder_name=sphinx_args["buildername"],
            doctree_dir=sphinx_args["doctreedir"],
            config_overrides=sphinx_args.get("confoverrides", {}),
            force_full_build=sphinx_args.get("freshenv", False),
            keep_going=sphinx_args.get("keep_going", False),
            num_jobs=sphinx_args.get("parallel", 1),
            quiet=sphinx_args.get("status", 1) is None,
            silent=sphinx_args.get("warning", 1) is None,
            tags=sphinx_args.get("tags", []),
            verbosity=sphinx_args.get("verbosity", 0),
            warning_is_error=sphinx_args.get("warningiserror", False),
        )

    def to_application_args(self, context: dict[str, Any]) -> dict[str, Any]:
        """Convert this into the equivalent Sphinx application arguments."""

        # On OSes like Fedora Silverblue, `/home` is symlinked to `/var/home`.  This
        # causes issues, since, depending on the origin of any path given to us `/home`
        # may or may not be the true location of the file - which introduces consistency
        # problems throughout the codebase.
        #
        # Resolving these paths here, should ensure that the agent always
        # reports the true location of any given directory.
        conf_dir = None
        if self.conf_dir is not None:
            conf_dir = pathlib.Path(self.conf_dir).resolve()
            self.conf_dir = str(conf_dir)

        # Resolve any config variables.
        #
        # This is a bit hacky, but let's go with it for now. The only config variable
        # we currently support is 'defaultBuildDir' which is derived from the value
        # of `conf_dir`. So we resolve the full path of `conf_dir` above, then resolve
        # the configuration variables here, before finally calling resolve() on the
        # remaining paths below.
        for name, value in dataclasses.asdict(self).items():
            # value can sometimes be Sphinx's `_StrPath` helper
            value = str(value)
            if (match := VARIABLE.match(value)) is None:
                continue

            replacement = self.resolve_config_variable(match.group(1), context)
            result = VARIABLE.sub(replacement.replace("\\", r"\\"), value)

            setattr(self, name, result)

        build_dir = pathlib.Path(self.build_dir).resolve()
        doctree_dir = pathlib.Path(self.doctree_dir).resolve()
        src_dir = pathlib.Path(self.src_dir).resolve()

        return {
            "buildername": self.builder_name,
            "confdir": str(conf_dir) if conf_dir is not None else None,
            "confoverrides": self.config_overrides,
            "doctreedir": str(doctree_dir),
            "freshenv": self.force_full_build,
            "keep_going": self.keep_going,
            "outdir": str(build_dir),
            "parallel": self.parallel,
            "srcdir": str(src_dir),
            "status": sys.stderr,
            "tags": self.tags,
            "verbosity": self.verbosity,
            "warning": sys.stderr,
            "warningiserror": self.warning_is_error,
        }

    def resolve_config_variable(self, name: str, context: dict[str, str]):
        """Resolve the value for the given configuration variable."""

        if name.lower() == "defaultbuilddir":
            if (cache_dir := context.get("cacheDir")) is None:
                raise RuntimeError(
                    f"Unable to resolve config variable {name!r}, "
                    "missing context value: 'cacheDir'"
                )

            dirname = self.conf_dir or self.src_dir
            project = hashlib.md5(dirname.encode()).hexdigest()  # noqa: S324
            return str(pathlib.Path(cache_dir, project))

        raise ValueError(f"Unknown configuration variable {name!r}")



from __future__ import annotations

import inspect
import logging
import os
import pathlib
import sys
import typing

from . import types
from .types import Uri
from .util import logger

if typing.TYPE_CHECKING:
    from types import ModuleType
    from typing import Any


DIAGNOSTIC_SEVERITY = {
    logging.ERROR: types.DiagnosticSeverity.Error,
    logging.INFO: types.DiagnosticSeverity.Information,
    logging.WARNING: types.DiagnosticSeverity.Warning,
}


class DiagnosticFilter(logging.Filter):
    """A logging handler that can extract errors from Sphinx's build output."""

    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.app = app

    def filter(self, record: logging.LogRecord) -> bool:
        conditions = [
            "sphinx" not in record.name,
            record.levelno not in {logging.WARNING, logging.ERROR},
        ]

        if any(conditions):
            return True

        loc = getattr(record, "location", "")
        uri, lineno = source_to_uri_and_linum(loc)

        if uri is None:
            conf = pathlib.Path(self.app.confdir, "conf.py")
            uri, lineno = (Uri.for_file(conf), None)

        line = lineno or 1

        try:
            # Not every message contains a string...
            if not isinstance(record.msg, str):
                message = str(record.msg)
            else:
                message = record.msg

            # Only attempt to format args if there are args to format
            if record.args is not None and len(record.args) > 0:
                message = message % record.args

        except Exception:
            message = str(record.msg)
            logger.error("Unable to format diagnostic message: %s", exc_info=True)

        diagnostic = types.Diagnostic(
            range=types.Range(
                start=types.Position(line=line - 1, character=0),
                end=types.Position(line=line, character=0),
            ),
            message=message,
            severity=DIAGNOSTIC_SEVERITY.get(
                record.levelno, types.DiagnosticSeverity.Warning
            ),
        )

        self.app.esbonio.diagnostics.setdefault(uri, set()).add(diagnostic)
        return True


def source_to_uri_and_linum(
    location: str | None,
) -> tuple[Uri | None, int | None]:
    """Convert the given source location to a uri and corresponding line number

    Parameters
    ----------
    location
       The location to convert

    Returns
    -------
    Tuple[Optional[Uri], Optional[int]]
       The corresponding uri and line number, if known
    """
    if location is None:
        return None, None

    lineno = None
    path, parts = _get_location_path(location)

    if len(parts) == 1:
        try:
            lineno = int(parts[0])
        except ValueError:
            pass

    if len(parts) == 2 and parts[0].startswith("docstring of "):
        target = parts[0].replace("docstring of ", "")
        lineno = _get_docstring_linum(target, parts[1])

    return Uri.for_file(path), lineno


def _get_location_path(location: str) -> tuple[str, list[str]]:
    """Determine the filepath from the given location."""

    if location.startswith("internal padding before "):
        location = location.replace("internal padding before ", "")

    if location.startswith("internal padding after "):
        location = location.replace("internal padding after ", "")

    path, *parts = location.split(":")

    # On windows the rest of the path will be the first element of parts
    if pathlib.Path(location).drive:
        path += f":{parts.pop(0)}"

    # Diagnostics in .. included:: files are reported relative to the process'
    # working directory, so ensure the path is absolute.
    path = os.path.abspath(path)

    return path, parts


def _get_docstring_linum(target: str, offset: str) -> int | None:
    # The containing module will be the longest substring we can find in target
    candidates = [m for m in sys.modules.keys() if target.startswith(m)] + [""]
    module = sys.modules.get(sorted(candidates, key=len, reverse=True)[0], None)

    if module is None:
        return None

    obj: ModuleType | Any | None = module
    dotted_name = target.replace(module.__name__ + ".", "")

    for name in dotted_name.split("."):
        obj = getattr(obj, name, None)
        if obj is None:
            return None

    try:
        _, line = inspect.getsourcelines(obj)  # type: ignore

        # Correct off by one error for docstrings that don't start with a newline.
        nl = (obj.__doc__ or "").startswith("\n")
        return line + int(offset) - (not nl)
    except Exception:
        logger.debug("Unable to determine diagnostic location\n%s", exc_info=True)
        return None



from __future__ import annotations

import enum
import json
import logging
import logging.config
import textwrap
from logging.handlers import MemoryHandler
from typing import Any

import attrs
from lsprotocol import types

from esbonio import server

LOG_LEVELS = {"CRITICAL", "FATAL", "ERROR", "WARN", "WARNING", "INFO", "DEBUG"}


class WindowLogMessageHandler(logging.Handler):
    """A logging handler that will send log records to an LSP client as
    ``window/logMessage`` notifications."""

    def __init__(
        self,
        server: server.EsbonioLanguageServer,
    ):
        super().__init__()
        self.server = server

    def emit(self, record: logging.LogRecord) -> None:
        """Sends the record to the client."""

        # To avoid infinite recursions, we can't process records coming from pygls.
        if "pygls" in record.name:
            return

        log = self.format(record).strip()
        self.server.window_log_message(
            types.LogMessageParams(message=log, type=types.MessageType.Log)
        )


@attrs.define
class LoggerConfiguration:
    """Configuration options for a given logger."""

    level: str | None = attrs.field(default=None)
    """The logging level to use, if not set the default logging level will be used."""

    format: str | None = attrs.field(default=None)
    """The log format to use, if not set the default logging level will be used."""

    filepath: str | None = attrs.field(default=None)
    """If set log to a file"""

    stderr: bool | None = attrs.field(default=None)
    """If True, log to stderr, if not set the default value will be used."""

    window: bool | None = attrs.field(default=None)
    """If True, send message as a ``window/logMessage`` notification, if not set the
    default value will be used"""


class LoggingConfigBuilder:
    """Helper class for converting the user's config into the logging config."""

    def __init__(self):
        self.formatters = {}
        self.handlers = {}
        self.loggers = {}

    def _get_formatter(self, format: str) -> str:
        """Return the name of the formatter with the given format string.

        If no such formatter exists, it will be created.
        """
        for key, config in self.formatters.items():
            if config["format"] == format:
                return key

        key = f"fmt{len(self.formatters) + 1:02d}"
        self.formatters[key] = dict(format=format)
        return key

    def _get_file_handler(self, filepath: str, formatter: str) -> str:
        """Return the name of the handler that will log to the given filepath using the
        given formatter name.

        If no such handler exists, it will be created.
        """
        for key, config in self.handlers.items():
            if config.get("class", None) != "logging.FileHandler":
                continue

            if (
                config.get("formatter", None) == formatter
                and config.get("filename", None) == filepath
            ):
                return key

        key = f"file{len(self.handlers) + 1:02d}"
        self.handlers[key] = {
            "class": "logging.FileHandler",
            "level": "DEBUG",  # this way we can handle logs from loggers at any level
            "formatter": formatter,
            "filename": filepath,
        }

        return key

    def _get_stderr_handler(self, formatter: str) -> str:
        """Return the name of the handler that will log to stderr using the given
        formatter name.

        If no such handler exists, it will be created.
        """
        for key, config in self.handlers.items():
            if config.get("class", None) != "logging.StreamHandler":
                continue

            if config.get("formatter", None) == formatter:
                return key

        key = f"stderr{len(self.handlers) + 1:02d}"
        self.handlers[key] = {
            "class": "logging.StreamHandler",
            "level": "DEBUG",  # this way we can handle logs from loggers at any level
            "formatter": formatter,
            "stream": "ext://sys.stderr",
        }

        return key

    def _get_window_handler(
        self, server: server.EsbonioLanguageServer, formatter: str
    ) -> str:
        """Return the name of the handler that will send messages as
        ``window/logMessage`` notificaitions, using the given formatter name.

        If no such formatter exists, it will be created.
        """
        handler_class = f"{__name__}.WindowLogMessageHandler"

        for key, config in self.handlers.items():
            if config.get("()", None) != handler_class:
                continue

            if config.get("formatter", None) == formatter:
                return key

        key = f"window{len(self.handlers) + 1:02d}"
        self.handlers[key] = {
            "()": handler_class,
            "level": "DEBUG",  # this way we can handle logs from loggers at any level
            "formatter": formatter,
            "server": server,
        }

        return key

    def add_logger(
        self,
        name: str,
        level: str,
        format: str,
        filepath: str | None,
        stderr: bool,
        window: server.EsbonioLanguageServer | None,
    ):
        """Add a configuration for the given logger

        Parameters
        ----------
        name
           The logger name to add a configuration for

        level
           The level at which to log messages

        format
           The format string to apply to messages

        filepath
           If set, record log messages in the given filepath

        stderr
           If ``True``, print messages from this logger to stderr

        window
           If set, send messages from this logger to the client as
           ``window/logMessage`` notifications via the given server instance
        """
        fmt = self._get_formatter(format)
        handlers = []

        if filepath:
            handlers.append(self._get_file_handler(filepath, fmt))

        if stderr:
            handlers.append(self._get_stderr_handler(fmt))

        if window:
            handlers.append(self._get_window_handler(window, fmt))

        if (level := level.upper()) not in LOG_LEVELS:
            level = "DEBUG"

        self.loggers[name] = dict(level=level, propagate=False, handlers=handlers)

    def finish(self) -> dict[str, Any]:
        """Return the final configuration."""
        return dict(
            version=1,
            disable_existing_loggers=True,
            formatters=self.formatters,
            handlers=self.handlers,
            loggers=self.loggers,
        )


@attrs.define
class LoggingConfig:
    """Configuration options for server logging."""

    level: str = attrs.field(default="error")
    """The default logging level."""

    format: str = attrs.field(default="[%(name)s] %(message)s")
    """The log format string to use."""

    filepath: str | None = attrs.field(default=None)
    """If set, log to a file by default"""

    stderr: bool = attrs.field(default=True)
    """If set, log to stderr by default"""

    window: bool = attrs.field(default=False)
    """If set, send message as a ``window/logMessage`` notification"""

    config: dict[str, LoggerConfiguration] = attrs.field(factory=dict)
    """Configuration of individual loggers"""

    show_deprecation_warnings: bool = attrs.field(default=False)
    """Developer flag to enable deprecation warnings."""

    def to_logging_config(self, server: server.EsbonioLanguageServer) -> dict[str, Any]:
        """Convert the user's config into a config dict that can be passed to the
        ``logging.config.dictConfig()`` function.

        Parameters
        ----------
        server
           The language server instance (required for ``window/logMessages``)
        """

        builder = LoggingConfigBuilder()

        # Ensure that there is at least an esbonio logger and a sphinx logger present in
        # the config.
        if "esbonio" not in self.config:
            builder.add_logger(
                "esbonio",
                self.level,
                self.format,
                filepath=self.filepath,
                stderr=self.stderr,
                window=server if self.window else None,
            )

        if "sphinx" not in self.config:
            builder.add_logger(
                "sphinx",
                "info",
                "%(message)s",
                filepath=self.filepath,
                stderr=self.stderr,
                window=server if self.window else None,
            )

        # Process any custom logger configuration
        for name, logger_config in self.config.items():
            window = (
                logger_config.window
                if logger_config.window is not None
                else self.window
            )
            builder.add_logger(
                name,
                logger_config.level or self.level,
                logger_config.format or self.format,
                filepath=(
                    logger_config.filepath
                    if logger_config.filepath is not None
                    else self.filepath
                ),
                stderr=(
                    logger_config.stderr
                    if logger_config.stderr is not None
                    else self.stderr
                ),
                window=server if window else None,
            )

        return builder.finish()


class LogManager(server.LanguageFeature):
    """Manages the logging setup for the server."""

    def initialized(self, params: types.InitializedParams):
        """Setup logging."""
        self.server.configuration.subscribe(
            "esbonio.logging", LoggingConfig, self.setup_logging
        )

    def setup_logging(self, event: server.ConfigChangeEvent[LoggingConfig]):
        """Setup logging according to the given config.

        Parameters
        ----------
        event
           The configuration change event
        """

        records = []
        if event.previous is None:
            # Messages received during initial startup are cached in a MemoryHandler
            # instance, let's resuce them so they can be replayed against the new config.
            for handler in logging.getLogger().handlers:
                if isinstance(handler, MemoryHandler):
                    records = handler.buffer

        config = event.value.to_logging_config(self.server)
        logging.config.dictConfig(config)

        # Replay any captured messages against the new config.
        for record in records:
            logger = logging.getLogger(record.name)
            if logger.isEnabledFor(record.levelno):
                logger.handle(record)


def dump(obj) -> str:
    """Debug helper function that converts an object to JSON."""

    def default(o):
        if isinstance(o, enum.Enum):
            return o.value

        fields = {}
        for k, v in o.__dict__.items():
            if v is None:
                continue

            # Truncate long strings - but not uris!
            if isinstance(v, str) and not k.lower().endswith("uri"):
                v = textwrap.shorten(v, width=25)

            fields[k] = v

        return fields

    return json.dumps(obj, default=default, indent=2)


def esbonio_setup(server: server.EsbonioLanguageServer):
    manager = LogManager(server)
    server.add_feature(manager)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from docutils import nodes
from sphinx.application import Sphinx

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Esbonio Demo"
copyright = "2023, Esbonio Developers"
author = "Esbonio Developers"
release = "1.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.intersphinx",
    "sphinx_design",
    "myst_parser",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "env", ".tox", "README.md"]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
    "myst": ("https://myst-parser.readthedocs.io/en/latest", None),
}


linkcheck_allowed_redirects = {
    # All HTTP redirections from the source URI to the canonical URI will be treated as "working".
    r"https://sphinx-doc\.org/.*": r"https://sphinx-doc\.org/en/master/.*"
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_title = "Esbonio Demo"
html_theme_options = {
    "source_repository": "https://github.com/swyddfa/esbonio",
    "source_branch": "develop",
    "source_directory": "lib/esbonio/tests/workspaces/demo/",
}


def lsp_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    """Link to sections within the lsp specification."""

    anchor = text.replace("/", "_")
    ref = f"https://microsoft.github.io/language-server-protocol/specifications/lsp/3.17/specification/#{anchor}"

    node = nodes.reference(rawtext, text, refuri=ref, **options)
    return [node], []


def setup(app: Sphinx):
    app.add_role("lsp", lsp_role)



from __future__ import annotations

import pytest
from lsprotocol import types
from pytest_lsp import LanguageClient

from esbonio.server.testing import range_from_str


def document_symbol(
    name: str,
    kind: types.SymbolKind,
    range: str,
    selection_range: str = "",
    children: list[types.DocumentSymbol] | None = None,
    detail: str = "",
) -> types.DocumentSymbol:
    """Helper for defining symbol instances."""
    return types.DocumentSymbol(
        name=name,
        kind=kind,
        range=range_from_str(range),
        selection_range=range_from_str(selection_range or range),
        children=children or None,
        detail=detail,
    )


@pytest.mark.parametrize(
    "filepath, expected",
    [
        (["conf.py"], None),
        (
            ["rst", "symbols.rst"],
            [
                document_symbol(
                    "Symbols",
                    types.SymbolKind.String,
                    "1:0-1:6",
                    children=[
                        document_symbol(
                            "What is a symbol?",
                            types.SymbolKind.Class,
                            "3:0-3:16",
                            detail="admonition",
                        ),
                        document_symbol(
                            "Document Symbols", types.SymbolKind.String, "14:0-14:15"
                        ),
                        document_symbol(
                            "Workspace Symbols", types.SymbolKind.String, "23:0-23:16"
                        ),
                    ],
                )
            ],
        ),
        (
            ["myst", "symbols.md"],
            [
                document_symbol(
                    "Symbols",
                    types.SymbolKind.String,
                    "0:0-0:6",
                    children=[
                        document_symbol(
                            "What is a symbol?",
                            types.SymbolKind.Class,
                            "2:0-2:16",
                            detail="admonition",
                        ),
                        document_symbol(
                            "Document Symbols",
                            types.SymbolKind.String,
                            "12:0-12:15",
                            children=[
                                document_symbol(
                                    "note",
                                    types.SymbolKind.Class,
                                    "20:0-20:3",
                                    detail="note",
                                ),
                            ],
                        ),
                        document_symbol(
                            "Workspace Symbols",
                            types.SymbolKind.String,
                            "27:0-27:16",
                        ),
                    ],
                )
            ],
        ),
    ],
)
@pytest.mark.asyncio(loop_scope="session")
async def test_document_symbols(
    client: LanguageClient,
    uri_for,
    filepath: list[str],
    expected: list[types.DocumentSymbol] | None,
):
    """Ensure that we handle ``textDocument/documentSymbols`` requests correctly."""

    test_uri = uri_for("workspaces", "demo", *filepath)
    actual = await client.text_document_document_symbol_async(
        types.DocumentSymbolParams(
            text_document=types.TextDocumentIdentifier(uri=str(test_uri))
        )
    )

    if expected is None:
        assert actual is None

    else:
        assert len(actual) == len(expected)
        for actual_symbol, expected_symbol in zip(actual, expected, strict=False):
            check_document_symbol(actual_symbol, expected_symbol)


@pytest.mark.parametrize(
    "query, expected",
    [
        # No query -> return all symbols.
        (
            "",
            {
                # (
                #   path (relative to workspace root),
                #   range (s:c-e:c),
                #   "name",
                #   types.SymbolKind,
                #   "container_name"
                #  )
                (
                    "/rst/symbols.rst",
                    "1:0-1:6",
                    "Symbols",
                    types.SymbolKind.String,
                    "",
                ),
                (
                    "/rst/symbols.rst",
                    "3:0-3:16",
                    "What is a symbol? admonition",
                    types.SymbolKind.Class,
                    "Symbols",
                ),
                (
                    "/myst/symbols.md",
                    "0:0-0:6",
                    "Symbols",
                    types.SymbolKind.String,
                    "",
                ),
                (
                    "/myst/symbols.md",
                    "2:0-2:16",
                    "What is a symbol? admonition",
                    types.SymbolKind.Class,
                    "Symbols",
                ),
                (
                    "/myst/symbols.md",
                    "20:0-20:3",
                    "note",
                    types.SymbolKind.Class,
                    "Document Symbols",
                ),
            },
        ),
        # We should be able to query by symbol name
        (
            "Symbols",
            {
                (
                    "/rst/symbols.rst",
                    "1:0-1:6",
                    "Symbols",
                    types.SymbolKind.String,
                    "",
                ),
                (
                    "/rst/symbols.rst",
                    "14:0-14:15",
                    "Document Symbols",
                    types.SymbolKind.String,
                    "Symbols",
                ),
                (
                    "/rst/symbols.rst",
                    "23:0-23:16",
                    "Workspace Symbols",
                    types.SymbolKind.String,
                    "Symbols",
                ),
                (
                    "/myst/symbols.md",
                    "0:0-0:6",
                    "Symbols",
                    types.SymbolKind.String,
                    "",
                ),
                (
                    "/myst/symbols.md",
                    "12:0-12:15",
                    "Document Symbols",
                    types.SymbolKind.String,
                    "Symbols",
                ),
                (
                    "/myst/symbols.md",
                    "27:0-27:16",
                    "Workspace Symbols",
                    types.SymbolKind.String,
                    "Symbols",
                ),
            },
        ),
        # We should also be able to query by (document) symbol `detail` e.g. a directive name
        (
            "admonition",
            {
                (
                    "/myst/symbols.md",
                    "2:0-2:16",
                    "What is a symbol? admonition",
                    types.SymbolKind.Class,
                    "Symbols",
                ),
                (
                    "/rst/symbols.rst",
                    "3:0-3:16",
                    "What is a symbol? admonition",
                    types.SymbolKind.Class,
                    "Symbols",
                ),
            },
        ),
        # Make sure we don't return anything when there are no matches
        ("--not-a-real-symbol-name--", None),
    ],
)
@pytest.mark.asyncio(loop_scope="session")
async def test_workspace_symbols(
    client: LanguageClient,
    query: str,
    uri_for,
    expected: set[tuple[str, str, str, types.SymbolKind, str]] | None,
):
    """Ensure that we handle ``workspace/symbol`` requests correctly."""

    workspace_uri = str(uri_for("workspaces", "demo"))
    result = await client.workspace_symbol_async(
        types.WorkspaceSymbolParams(query=query)
    )

    if expected is None:
        assert result is None
        return

    assert result is not None

    actual = set()
    for symbol in result:
        loc = symbol.location
        uri = loc.uri.replace(workspace_uri, "")

        actual.add(
            (uri, repr(loc.range), symbol.name, symbol.kind, symbol.container_name),
        )

    for symbol in expected:
        assert symbol in actual


def check_document_symbol(actual: types.DocumentSymbol, expected: types.DocumentSymbol):
    """Ensure that the given ``DocumentSymbols`` are equivalent."""

    assert isinstance(actual, types.DocumentSymbol)

    assert actual.name == expected.name
    assert actual.kind == expected.kind
    assert actual.range == expected.range
    assert actual.selection_range == expected.selection_range

    if expected.children is None:
        assert actual.children is None
        return

    assert actual.children is not None
    assert len(actual.children) == len(expected.children), (
        f"Children mismatch in symbol '{actual.name}'"
    )

    for actual_child, expected_child in zip(
        actual.children, expected.children, strict=False
    ):
        check_document_symbol(actual_child, expected_child)



from __future__ import annotations

import json
import logging
import os
import pathlib
import typing

from lsprotocol import types as lsp

from esbonio import server
from esbonio.server.features import roles
from esbonio.server.features.project_manager import ProjectManager
from esbonio.sphinx_agent import types

if typing.TYPE_CHECKING:
    import cattrs

    from esbonio.server import Uri
    from esbonio.server.features.project_manager import Project


TARGET_KINDS = {
    "attribute": lsp.CompletionItemKind.Field,
    "doc": lsp.CompletionItemKind.File,
    "class": lsp.CompletionItemKind.Class,
    "envvar": lsp.CompletionItemKind.Variable,
    "function": lsp.CompletionItemKind.Function,
    "method": lsp.CompletionItemKind.Method,
    "module": lsp.CompletionItemKind.Module,
    "term": lsp.CompletionItemKind.Text,
}


class ObjectsProvider(roles.RoleTargetProvider):
    """Expose domain objects as potential role targets"""

    def __init__(
        self,
        logger: logging.Logger,
        converter: cattrs.Converter,
        manager: ProjectManager,
    ):
        self.manager = manager
        self.logger = logger
        self.converter = converter

    async def find_target_definition(  # type: ignore
        self,
        context: server.DefinitionContext,
        target: str,
        *,
        obj_types: list[str] | None,
        projects: list[str] | None,
        **kwargs,
    ) -> list[lsp.Location] | None:
        """Find the definition(s) for the given role target."""
        if obj_types is None:
            self.logger.debug("Unable to find definitions, missing object types!")
            return None

        if projects is not None:
            # does not make sense for intersphinx targets...
            return None

        if (project := self.manager.get_project(context.uri)) is None:
            return None

        self.logger.debug("%r, %r, %r, %r", context, target, obj_types, projects)
        db = await project.get_db()
        query = (
            "SELECT "  # noqa: S608
            "  location "
            "FROM objects "
            f'WHERE printf("%s:%s", objects.domain, objects.objtype) in ({", ".join("?" for _ in obj_types)})'
            "       AND objects.name = ?"
        )

        # Hack for absolute docnames...
        if "std:doc" in obj_types and target.startswith("/"):
            target = target[1:]

        cursor = await db.execute(query, (*obj_types, target))
        if (result := await cursor.fetchall()) is None:
            return None

        locations: list[lsp.Location] = []

        for item, *_ in result:
            if item is None:
                continue

            try:
                obj = json.loads(item)
                locations.append(self.converter.structure(obj, lsp.Location))
            except Exception:
                self.logger.exception(
                    "Unable to construct Location instance from value: %r", item
                )

        return locations

    async def hover_target(  # type: ignore
        self,
        context: server.HoverContext,
        target: str,
        *,
        obj_types: list[str] | None,
        projects: list[str] | None,
        **kwargs,
    ) -> str | None:
        """Find the hover text for the given role target."""
        if obj_types is None:
            self.logger.debug("Unable to find hover text, missing object types!")
            return None

        if (project := self.manager.get_project(context.uri)) is None:
            return None

        self.logger.debug("%r, %r, %r, %r", context, target, obj_types, projects)
        db = await project.get_db()
        query = (
            "SELECT "  # noqa: S608
            "  description "
            "FROM objects "
            f'WHERE printf("%s:%s", objects.domain, objects.objtype) in ({", ".join("?" for _ in obj_types)})'
            "       AND objects.name = ?"
        )

        # Hack for absolute docnames...
        if "std:doc" in obj_types and target.startswith("/"):
            target = target[1:]

        cursor = await db.execute(query, (*obj_types, target))
        if (result := await cursor.fetchall()) is None:
            return None

        for item, *_ in result:
            if item is None:
                continue

            return item

        return None

    async def resolve_target_link(
        self,
        context: server.DocumentLinkContext,
        target: str,
        *,
        obj_types: list[str] | None = None,
        projects: list[str] | None = None,
        **kwargs,
    ) -> None | str | tuple[str, str | None]:
        if obj_types is None:
            self.logger.debug("Unable to resolve link, missing object types!")
            return None

        self.logger.debug("%s %s %s %s", context, target, obj_types, projects)
        if projects is not None:
            return await self._resolve_intersphinx_link(
                context, target, obj_types, projects
            )

        if projects is None and "std:doc" in obj_types:
            # Other roles like :ref: do not make sense as the ``textDocument/documentLink``
            # api doesn't support specific locations in a file like goto definition does.
            return await self._resolve_doc_link(context, target)

        return None

    async def _resolve_intersphinx_link(
        self,
        context: server.DocumentLinkContext,
        target: str,
        obj_types: list[str],
        projects: list[str],
    ):
        """Resolve ``textDocument/documentLink`` requests for intersphinx references."""

        if (project := self.manager.get_project(context.uri)) is None:
            return None

        db = await project.get_db()
        query = (
            "SELECT "  # noqa: S608
            '  printf("%s%s", intersphinx_projects.uri, objects.docname) as uri,'
            '  printf("%s v%s", intersphinx_projects.name, intersphinx_projects.version) as source,'
            "  objects.name,"
            "  objects.display "
            "FROM objects JOIN intersphinx_projects "
            "ON objects.project = intersphinx_projects.id "
            f"WHERE objects.project in ({', '.join('?' for _ in projects)}) "
            f'  AND printf("%s:%s", objects.domain, objects.objtype) in ({", ".join("?" for _ in obj_types)})'
            "   AND objects.name = ?"
        )

        cursor = await db.execute(query, (*projects, *obj_types, target))
        if (result := await cursor.fetchone()) is None:
            return None

        uri, source, name, display = result
        display = name if display == "-" else display

        return uri, f"{display} - {source}"

    async def _resolve_doc_link(
        self, context: server.UriContext, target: str
    ) -> str | None:
        """Resolve ``textDocument/documentLink`` requests for local ``:doc:`` references."""

        if (project := self.manager.get_project(context.uri)) is None:
            return None

        if target.startswith("/"):
            docname = target[1:]

        elif (result := await project.uri_to_docname(context.uri)) is None:
            self.logger.debug("Unable to find docname for uri: %r", context.uri)
            return None

        else:
            docname = os.path.normpath(pathlib.Path(result).parent / target)

        return await project.docname_to_uri(docname)

    async def suggest_targets(
        self,
        context: server.CompletionContext,
        *,
        obj_types: list[str] | None = None,
        projects: list[str] | None = None,
        **kwargs,
    ) -> list[lsp.CompletionItem] | None:
        #  TODO: Handle .. currentmodule

        if obj_types is None:
            self.logger.debug("Unable to suggest targets, missing object types!")
            return None

        if (project := self.manager.get_project(context.uri)) is None:
            return None

        items = []
        db = await project.get_db()

        query, parameters = self._prepare_target_query(projects, obj_types)
        cursor = await db.execute(query, parameters)

        for name, display, type_ in await cursor.fetchall():
            kind = TARGET_KINDS.get(type_, lsp.CompletionItemKind.Reference)

            insert_text = None
            if type_ == "doc":
                # Insert an absolute reference, this way the suggestion is always correct.
                insert_text = f"/{name}"

            items.append(
                lsp.CompletionItem(
                    label=name,
                    detail=None if display == "-" else display,
                    kind=kind,
                    insert_text=insert_text,
                ),
            )

        return items

    def _prepare_target_query(self, projects: list[str] | None, obj_types: list[str]):
        """Prepare the query to use when looking up targets."""

        select = "SELECT name, display, objtype FROM objects"
        where = []
        parameters = []

        if projects is None:
            self.logger.debug(
                "Suggesting targets from the local project for types: %s", obj_types
            )
            where.append("project IS NULL")

        else:
            self.logger.debug(
                "Suggesting targets from projects %s for types: %s", projects, obj_types
            )

            placeholders = ", ".join("?" for _ in range(len(projects)))
            where.append(f"project IN ({placeholders})")
            parameters.extend(projects)

        placeholders = ", ".join("?" for _ in range(len(obj_types)))
        where.append(f"printf('%s:%s', domain, objtype) IN ({placeholders})")
        parameters.extend(obj_types)

        query = " ".join([select, "WHERE", " AND ".join(where)])

        return query, tuple(parameters)


class SphinxRoles(roles.RoleProvider):
    """Support for roles in a sphinx project."""

    def __init__(self, manager: ProjectManager):
        self.manager = manager

    async def get_default_domain(self, project: Project, uri: Uri) -> str:
        """Get the name of the default domain for the given document"""

        # Does the document have a default domain set?
        results = await project.find_symbols(
            uri=str(uri.resolve()),
            kind=lsp.SymbolKind.Class.value,
            detail="default-domain",
        )
        if len(results) > 0:
            default_domain = results[0][1]
        else:
            default_domain = None

        primary_domain = await project.get_config_value("primary_domain")
        return default_domain or primary_domain or "py"

    async def get_role(self, uri: Uri, name: str) -> types.Role | None:
        """Return the role with the given name."""

        if (project := self.manager.get_project(uri)) is None:
            return None

        if (role := await project.get_role(name)) is not None:
            return role

        if (role := await project.get_role(f"std:{name}")) is not None:
            return role

        default_domain = await self.get_default_domain(project, uri)
        return await project.get_role(f"{default_domain}:{name}")

    async def suggest_roles(
        self, context: server.CompletionContext
    ) -> list[types.Role] | None:
        """Given a completion context, suggest roles that may be used."""

        if (project := self.manager.get_project(context.uri)) is None:
            return None

        default_domain = await self.get_default_domain(project, context.uri)

        result: list[types.Role] = []
        for name, implementation in await project.get_roles():
            # std: directives can be used unqualified
            if name.startswith("std:"):
                short_name = name.replace("std:", "")
                result.append(
                    types.Role(name=short_name, implementation=implementation)
                )

            # Also suggest unqualified versions of directives from the currently active domain.
            elif name.startswith(f"{default_domain}:"):
                short_name = name.replace(f"{default_domain}:", "")
                result.append(
                    types.Role(name=short_name, implementation=implementation)
                )

            result.append(types.Role(name=name, implementation=implementation))

        return result


def esbonio_setup(
    esbonio: server.EsbonioLanguageServer,
    project_manager: ProjectManager,
    roles_feature: roles.RolesFeature,
):
    role_provider = SphinxRoles(project_manager)
    obj_provider = ObjectsProvider(
        esbonio.logger.getChild("ObjectsProvider"), esbonio.converter, project_manager
    )

    roles_feature.add_role_provider(role_provider)
    roles_feature.add_role_target_provider("objects", obj_provider)



from __future__ import annotations

import inspect
import typing

from lsprotocol import types as lsp

from esbonio import server
from esbonio.sphinx_agent import types

from . import providers
from .providers import DirectiveArgumentProvider

if typing.TYPE_CHECKING:
    from collections.abc import Coroutine
    from typing import Any

    from esbonio.server import Uri


class DirectiveProvider:
    """Base class for directive providers"""

    def get_directive(
        self, uri: Uri, name: str
    ) -> types.Directive | None | Coroutine[Any, Any, types.Directive | None]:
        """Return the definition of the given directive, if known

        Parameters
        ----------
        uri
           The uri of the document in which the directive name appears

        name
           The name of the directive, as the user would type in a document
        """
        return None

    def suggest_directives(
        self, context: server.CompletionContext
    ) -> (
        list[types.Directive] | None | Coroutine[Any, Any, list[types.Directive] | None]
    ):
        """Given a completion context, suggest directives that may be used."""
        return None


class DirectiveFeature(server.LanguageFeature):
    """'Backend' support for directives.

    It's this language feature's responsibility to provide an API that exposes the
    information a "frontend" language feature may want.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._directive_providers: dict[int, DirectiveProvider] = {}
        self._argument_providers: dict[str, DirectiveArgumentProvider] = {}

    def add_directive_provider(self, provider: DirectiveProvider):
        """Register a directive provider.

        Parameters
        ----------
        provider
           The directive provider
        """
        self._directive_providers[id(provider)] = provider

    def add_directive_argument_provider(
        self, name: str, provider: DirectiveArgumentProvider
    ):
        """Register a directive argument provider.

        Parameters
        ----------
        provider
           The directive argument provider
        """
        if (existing := self._argument_providers.get(name)) is not None:
            raise ValueError(
                f"DirectiveArgumentProvider {provider!r} conflicts with existing "
                f"provider: {existing!r}"
            )

        self._argument_providers[name] = provider

    async def suggest_directives(
        self, context: server.CompletionContext
    ) -> list[types.Directive]:
        """Suggest directives that may be used, given a completion context.

        Parameters
        ----------
        context
           The completion context.
        """
        items: list[types.Directive] = []

        for provider in self._directive_providers.values():
            try:
                result: list[types.Directive] | None = None

                aresult = provider.suggest_directives(context)
                if inspect.isawaitable(aresult):
                    result = await aresult

                if result:
                    items.extend(result)
            except Exception:
                name = type(provider).__name__
                self.logger.error(
                    "Error in '%s.suggest_directives'", name, exc_info=True
                )

        return items

    async def get_directive(self, uri: Uri, name: str) -> types.Directive | None:
        """Return the definition of the given directive name.

        Parameters
        ----------
        uri
           The uri of the document in which the directive name appears

        name
           The name of the directive, as the user would type into a document.

        Returns
        -------
        types.Directive | None
           The directive's definition, if known
        """
        for provider in self._directive_providers.values():
            try:
                result: types.Directive | None = None

                aresult = provider.get_directive(uri, name)
                if inspect.isawaitable(aresult):
                    result = await aresult

                if result is not None:
                    return result
            except Exception:
                name = type(provider).__name__
                self.logger.error("Error in '%s.get_directive'", name, exc_info=True)

        return None

    async def find_argument_definition(
        self, context: server.DefinitionContext, directive_name: str, argument: str
    ) -> list[lsp.Location] | None:
        """Find the definition of the directive's argument.

        Parameters
        ----------
        context
           The definition context

        directive_name
           The directive to find the argument definition for

        argument
           The directive's argument

        Returns
        -------
        list[lsp.Location] | None
           The argument's defintion(s), if known
        """
        if (directive := await self.get_directive(context.uri, directive_name)) is None:
            self.logger.debug("Unknown directive '%s'", directive_name)
            return None

        if not directive.argument_providers:
            return None

        self.logger.debug(
            "Finding argument defintion for directive: '%s' (%s)",
            directive.name,
            directive.implementation,
        )

        for spec in directive.argument_providers:
            if (provider := self._argument_providers.get(spec.name)) is None:
                self.logger.error("Unknown argument provider: '%s'", spec.name)
                continue

            try:
                result = provider.find_argument_definition(
                    context, argument, **spec.kwargs
                )
                if inspect.isawaitable(result):
                    result = await result

                if result is not None:
                    return result

            except Exception:
                name = type(provider).__name__
                self.logger.exception("Error in '%s.find_argument_definition'", name)

        return None

    async def resolve_argument_link(
        self, context: server.DocumentLinkContext, directive_name: str, argument: str
    ) -> None | str | tuple[str, str | None]:
        """Resolve the link to the given directive argument, if possible.

        Parameters
        ----------
        context
           The document link context

        directive_name
           The directive to resolve the argument for

        argument
           The directive's argument
        """
        if (directive := await self.get_directive(context.uri, directive_name)) is None:
            self.logger.debug("Unknown directive '%s'", directive_name)
            return None

        if not directive.argument_providers:
            return None

        self.logger.debug(
            "Resolving argument link for directive: '%s' (%s)",
            directive.name,
            directive.implementation,
        )

        for spec in directive.argument_providers:
            if (provider := self._argument_providers.get(spec.name)) is None:
                self.logger.error("Unknown argument provider: '%s'", spec.name)
                continue

            try:
                result: None | str | tuple[str, str | None] = None

                aresult = provider.resolve_argument_link(
                    context, argument, **spec.kwargs
                )
                if inspect.isawaitable(aresult):
                    result = await aresult
                else:
                    result = aresult

                if result is not None:
                    return result

            except Exception:
                name = type(provider).__name__
                self.logger.error(
                    "Error in '%s.resolve_argument_link'", name, exc_info=True
                )

        return None

    async def suggest_arguments(
        self, context: server.CompletionContext, directive_name: str
    ) -> list[lsp.CompletionItem]:
        """Suggest directive arguments that may be used, given a completion context.

        Parameters
        ----------
        context
           The completion context

        directive_name
           The directive to suggest arguments for
        """
        if (directive := await self.get_directive(context.uri, directive_name)) is None:
            self.logger.debug("Unknown directive '%s'", directive_name)
            return []

        arguments = []
        self.logger.debug(
            "Suggesting arguments for directive: '%s' (%s)",
            directive.name,
            directive.implementation,
        )

        for spec in directive.argument_providers:
            if (provider := self._argument_providers.get(spec.name)) is None:
                self.logger.error("Unknown argument provider: '%s'", spec.name)
                continue

            try:
                result: list[lsp.CompletionItem] | None = None

                aresult = provider.suggest_arguments(context, **spec.kwargs)
                if inspect.isawaitable(aresult):
                    result = await aresult
                else:
                    result = aresult

                if result is not None:
                    arguments.extend(result)

            except Exception:
                name = type(provider).__name__
                self.logger.error(
                    "Error in '%s.suggest_arguments'", name, exc_info=True
                )

        return arguments


def esbonio_setup(server: server.EsbonioLanguageServer):
    directives = DirectiveFeature(server)
    directives.add_directive_argument_provider(
        "filepath", providers.FilepathProvider(server)
    )
    directives.add_directive_argument_provider(
        "values", providers.ValuesProvider(server)
    )

    server.add_feature(directives)



"""Helpers and utilities for writing tests."""

from lsprotocol import types


def range_from_str(spec: str) -> types.Range:
    """Create a range from the given string ``a:b-x:y``"""
    start, end = spec.split("-")
    sl, sc = start.split(":")
    el, ec = end.split(":")

    return types.Range(
        start=types.Position(line=int(sl), character=int(sc)),
        end=types.Position(line=int(el), character=int(ec)),
    )



import asyncio
import pathlib
import sys

import pytest_lsp
from lsprotocol import types
from pytest_lsp import ClientServerConfig
from pytest_lsp import LanguageClient
from pytest_lsp import client_capabilities

# Disable the sphinx integration - at least for now...
SERVER_CMD = [
    "-m",
    "esbonio.server",
    "--exclude",
    "esbonio.server.features.sphinx_manager",
]
TEST_DIR = pathlib.Path(__file__).parent.parent


@pytest_lsp.fixture(
    scope="session",
    params=["visual_studio_code", "neovim"],
    config=ClientServerConfig(
        server_command=[sys.executable, *SERVER_CMD],
    ),
)
async def client(request, uri_for, lsp_client: LanguageClient):
    workspace_uri = uri_for("sphinx-default", "workspace")

    await lsp_client.initialize_session(
        types.InitializeParams(
            capabilities=client_capabilities(request.param),
            initialization_options={"server": {"logLevel": "debug"}},
            workspace_folders=[
                types.WorkspaceFolder(uri=str(workspace_uri), name="sphinx-default"),
            ],
        )
    )

    yield

    # Teardown
    await asyncio.wait_for(lsp_client.shutdown_session(), timeout=2.0)



import pytest

from esbonio.server.features.project_manager import Project
from esbonio.server.features.sphinx_manager.client_subprocess import (
    SubprocessSphinxClient,
)


def anuri(base, *args):
    uri = base
    for a in args:
        uri /= a

    return str(uri.resolve())


@pytest.mark.asyncio
async def test_files_table(client: SubprocessSphinxClient, project: Project):
    """Ensure that we can correctly index all the files in the Sphinx project."""

    src = client.src_uri

    db = await project.get_db()
    cursor = await db.execute("SELECT * FROM files")
    results = await cursor.fetchall()
    actual = {r for r in results if "badfile" not in r[1]}

    expected = {
        (anuri(src, "index.rst"), "index", "index.html"),
        (anuri(src, "rst", "roles.rst"), "rst/roles", "rst/roles.html"),
        (anuri(src, "rst", "directives.rst"), "rst/directives", "rst/directives.html"),
        (
            anuri(src, "rst", "diagnostics.rst"),
            "rst/diagnostics",
            "rst/diagnostics.html",
        ),
        (
            anuri(src, "rst", "domains.rst"),
            "rst/domains",
            "rst/domains.html",
        ),
        (
            anuri(src, "rst", "domains", "python.rst"),
            "rst/domains/python",
            "rst/domains/python.html",
        ),
        (anuri(src, "rst", "symbols.rst"), "rst/symbols", "rst/symbols.html"),
        (anuri(src, "myst", "roles.md"), "myst/roles", "myst/roles.html"),
        (
            anuri(src, "myst", "directives.md"),
            "myst/directives",
            "myst/directives.html",
        ),
        (
            anuri(src, "myst", "diagnostics.md"),
            "myst/diagnostics",
            "myst/diagnostics.html",
        ),
        (anuri(src, "myst", "symbols.md"), "myst/symbols", "myst/symbols.html"),
        (anuri(src, "demo_rst.rst"), "demo_rst", "demo_rst.html"),
        (anuri(src, "demo_myst.md"), "demo_myst", "demo_myst.html"),
    }

    assert expected == actual



import logging
import pathlib
import sys

import pytest
import pytest_asyncio
from lsprotocol.types import WorkspaceFolder
from pygls.protocol import default_converter
from pygls.workspace import Workspace

from esbonio.server.features.project_manager import Project
from esbonio.server.features.sphinx_manager.client import ClientState
from esbonio.server.features.sphinx_manager.client_subprocess import (
    SubprocessSphinxClient,
)
from esbonio.server.features.sphinx_manager.client_subprocess import (
    make_test_sphinx_client,
)
from esbonio.server.features.sphinx_manager.config import SphinxConfig
from esbonio.server.features.sphinx_manager.config import SubProcess
from esbonio.server.features.sphinx_manager.config import get_module_path
from esbonio.sphinx_agent.app import Sphinx

logger = logging.getLogger(__name__)


@pytest.fixture
def build_dir(tmp_path_factory):
    _dir = tmp_path_factory.mktemp("build")
    print(f"Using build dir: {_dir}")

    return _dir


@pytest_asyncio.fixture
async def client(request, uri_for, build_dir):
    demo_workspace = uri_for("workspaces", "demo")
    test_uri = demo_workspace / "index.rst"

    workspace = Workspace(
        None,
        workspace_folders=[
            WorkspaceFolder(uri=str(demo_workspace), name="demo"),
        ],
    )
    config = SphinxConfig(
        enable_dev_tools=request.config.getoption("enable_devtools"),
        python_command=SubProcess(command=[sys.executable]),
        build_command=[
            "sphinx-build",
            "-M",
            "html",
            demo_workspace.fs_path,
            str(build_dir),
        ],
    )
    resolved = config.resolve(test_uri, workspace, logger)
    assert resolved is not None

    sphinx_client = await make_test_sphinx_client(resolved)
    assert sphinx_client.state == ClientState.Running

    await sphinx_client.build()
    yield sphinx_client

    await sphinx_client.stop()


@pytest.fixture
def app(client, build_dir):
    """Sphinx application instance, used for validating results."""

    # In order to load the pickled envrionment correctly, we need to temporarily put
    # the parent directory of the `sphinx_agent` module on the path.
    path = get_module_path("esbonio.sphinx_agent")
    assert path is not None

    sys.path.insert(0, str(path))

    _app = Sphinx(
        srcdir=client.sphinx_info.src_dir,
        confdir=client.sphinx_info.conf_dir,
        outdir=str(pathlib.Path(build_dir, "html")),
        doctreedir=str(pathlib.Path(build_dir, "doctrees")),
        buildername="html",
        confoverrides={
            "html_theme": "alabaster",
            "html_theme_options": {},
        },
    )

    sys.path.pop(0)

    return _app


@pytest_asyncio.fixture
async def project(client: SubprocessSphinxClient):
    """The Sphinx project as captured by the database created by the Sphinx agent."""
    project = Project(client.db, default_converter())

    yield project
    await project.close()
